from inspect import signature
from typing import Any, Optional

from fastapi import HTTPException, Request
from pydantic import BaseModel
from pymongo import timeout as query_timeout
from pymongo.errors import NetworkTimeout, PyMongoError

from maggma.api.models import Meta, Response
from maggma.api.query_operator import PaginationQuery, QueryOperator, SparseFieldsQuery
from maggma.api.resource import Resource
from maggma.api.resource.utils import attach_query_ops, generate_query_pipeline
from maggma.api.utils import STORE_PARAMS, merge_queries
from maggma.core import Store
from maggma.stores import S3Store


class PostOnlyResource(Resource):
    """
    Implements a REST Compatible Resource as a POST URL endpoint.
    """

    def __init__(
        self,
        store: Store,
        model: type[BaseModel],
        tags: Optional[list[str]] = None,
        query_operators: Optional[list[QueryOperator]] = None,
        key_fields: Optional[list[str]] = None,
        query: Optional[dict] = None,
        timeout: Optional[int] = None,
        include_in_schema: Optional[bool] = True,
        sub_path: Optional[str] = "/",
    ):
        """
        Args:
            store: The Maggma Store to get data from
            model: The pydantic model this Resource represents
            tags: List of tags for the Endpoint
            query_operators: Operators for the query language
            key_fields: List of fields to always project. Default uses SparseFieldsQuery
                to allow user to define these on-the-fly.
            timeout: Time in seconds Pymongo should wait when querying MongoDB
                before raising a timeout error
            include_in_schema: Whether the endpoint should be shown in the documented schema.
            sub_path: sub-URL path for the resource.
        """
        self.store = store
        self.tags = tags or []
        self.query = query or {}
        self.key_fields = key_fields
        self.versioned = False
        self.timeout = timeout

        self.include_in_schema = include_in_schema
        self.sub_path = sub_path
        self.response_model = Response[model]  # type: ignore

        self.query_operators = (
            query_operators
            if query_operators is not None
            else [
                PaginationQuery(),
                SparseFieldsQuery(
                    model,
                    default_fields=[self.store.key, self.store.last_updated_field],
                ),
            ]
        )

        super().__init__(model)

    def prepare_endpoint(self):
        """
        Internal method to prepare the endpoint by setting up default handlers
        for routes.
        """
        self.build_dynamic_model_search()

    def build_dynamic_model_search(self):
        model_name = self.model.__name__

        def search(**queries: dict[str, STORE_PARAMS]) -> dict:
            request: Request = queries.pop("request")  # type: ignore
            queries.pop("temp_response")  # type: ignore

            query_params = [
                entry for _, i in enumerate(self.query_operators) for entry in signature(i.query).parameters
            ]

            overlap = [key for key in request.query_params if key not in query_params]
            if any(overlap):
                raise HTTPException(
                    status_code=400,
                    detail="Request contains query parameters which cannot be used: {}".format(", ".join(overlap)),
                )

            query: dict[Any, Any] = merge_queries(list(queries.values()))  # type: ignore
            query["criteria"].update(self.query)

            self.store.connect()

            try:
                with query_timeout(self.timeout):
                    count = self.store.count(  # type: ignore
                        **{field: query[field] for field in query if field in ["criteria", "hint"]}
                    )

                    if isinstance(self.store, S3Store):
                        data = list(self.store.query(**query))  # type: ignore
                    else:
                        pipeline = generate_query_pipeline(query, self.store)

                        data = list(
                            self.store._collection.aggregate(
                                pipeline,
                                **{field: query[field] for field in query if field in ["hint"]},
                            )
                        )
            except (NetworkTimeout, PyMongoError) as e:
                if e.timeout:
                    raise HTTPException(
                        status_code=504,
                        detail="Server timed out trying to obtain data. Try again with a smaller request.",
                    )
                else:
                    raise HTTPException(
                        status_code=500,
                        detail="Server timed out trying to obtain data. Try again with a smaller request, "
                        "or remove sorting fields and sort data locally.",
                    )

            operator_meta = {}

            for operator in self.query_operators:
                data = operator.post_process(data, query)
                operator_meta.update(operator.meta())

            meta = Meta(total_doc=count)
            return {"data": data, "meta": {**meta.dict(), **operator_meta}}

        self.router.post(
            self.sub_path,
            tags=self.tags,
            summary=f"Post {model_name} documents",
            response_model=self.response_model,
            response_description=f"Post {model_name} data",
            response_model_exclude_unset=True,
        )(attach_query_ops(search, self.query_operators))



#!/usr/bin/env python
# coding utf-8

from asyncio import BoundedSemaphore, Queue, gather, get_event_loop
from concurrent.futures import ProcessPoolExecutor
from logging import getLogger
from types import GeneratorType
from typing import Any, Callable, Optional

from aioitertools import enumerate
from tqdm.auto import tqdm

from maggma.utils import primed

logger = getLogger("MultiProcessor")


class BackPressure:
    """
    Wrapper for an iterator to provide
    async access with backpressure.
    """

    def __init__(self, iterator, n):
        self.iterator = iter(iterator)
        self.back_pressure = BoundedSemaphore(n)

    def __aiter__(self):
        return self

    async def __anext__(self):
        await self.back_pressure.acquire()

        try:
            return next(self.iterator)
        except StopIteration:
            raise StopAsyncIteration

    async def release(self, async_iterator):
        """
        release iterator to pipeline the backpressure.
        """
        async for item in async_iterator:
            try:
                self.back_pressure.release()
            except ValueError:
                pass

            yield item


class AsyncUnorderedMap:
    """
    Async iterator that maps a function to an async iterator
    using an executor and returns items as they are done
    This does not guarantee order.
    """

    def __init__(self, func, async_iterator, executor):
        self.iterator = async_iterator
        self.func = func
        self.executor = executor

        loop = get_event_loop()

        self.fill_task = loop.create_task(self.get_from_iterator())

        self.done_sentinel = object()
        self.results = Queue()
        self.tasks = {}

    async def process_and_release(self, idx):
        future = self.tasks[idx]
        try:
            item = await future
            self.results.put_nowait(item)
        except Exception:
            pass
        finally:
            self.tasks.pop(idx)

    async def get_from_iterator(self):
        loop = get_event_loop()
        async for idx, item in enumerate(self.iterator):
            future = loop.run_in_executor(self.executor, safe_dispatch, (self.func, item))

            self.tasks[idx] = future
            # TODO - line below raises RUF006 error. Unsure about the best way to
            # resolve. See https://docs.astral.sh/ruff/rules/asyncio-dangling-task/
            loop.create_task(self.process_and_release(idx))  # noqa: RUF006

        await gather(*self.tasks.values())
        self.results.put_nowait(self.done_sentinel)

    def __aiter__(self):
        return self

    async def __anext__(self):
        item = await self.results.get()

        if item == self.done_sentinel:
            raise StopAsyncIteration

        return item


async def atqdm(async_iterator, *args, **kwargs):
    """
    Wrapper around tqdm for async generators.
    """
    _tqdm = tqdm(*args, **kwargs)
    async for item in async_iterator:
        _tqdm.update()
        yield item

    _tqdm.close()


async def grouper(async_iterator, n: int):
    """
    Collect data into fixed-length chunks or blocks.
    >>> list(grouper(3, 'ABCDEFG'))
    [['A', 'B', 'C'], ['D', 'E', 'F'], ['G']].

    Updated from:
    https://stackoverflow.com/questions/31164731/python-chunking-csv-file-multiproccessing/31170795#31170795

    Modified for async
    """
    chunk = []
    async for item in async_iterator:
        chunk.append(item)
        if len(chunk) >= n:
            yield chunk
            chunk.clear()
    if chunk != []:
        yield chunk


def safe_dispatch(val):
    func, item = val
    try:
        return func(item)
    except Exception as e:
        logger.error(e)
        return None


async def multi(
    builder,
    num_processes,
    no_bars=False,
    heartbeat_func: Optional[Callable[..., Any]] = None,
    heartbeat_func_kwargs: Optional[dict[Any, Any]] = None,
):
    builder.connect()
    cursor = builder.get_items()
    executor = ProcessPoolExecutor(num_processes)

    # Gets the total number of items to process by priming
    # the cursor
    total = None

    if isinstance(cursor, GeneratorType):
        try:
            cursor = primed(cursor)
            if hasattr(builder, "total"):
                total = builder.total
        except StopIteration:
            pass

    elif hasattr(cursor, "__len__"):
        total = len(cursor)
    elif hasattr(cursor, "count"):
        total = cursor.count()

    logger.info(
        f"Starting multiprocessing: {builder.__class__.__name__}",
        extra={
            "maggma": {
                "event": "BUILD_STARTED",
                "total": total,
                "builder": builder.__class__.__name__,
                "sources": [source.name for source in builder.sources],
                "targets": [target.name for target in builder.targets],
            }
        },
    )

    back_pressured_get = BackPressure(
        iterator=tqdm(cursor, desc="Get", total=total, disable=no_bars),
        n=builder.chunk_size,
    )

    processed_items = atqdm(
        async_iterator=AsyncUnorderedMap(
            func=builder.process_item,
            async_iterator=back_pressured_get,
            executor=executor,
        ),
        total=total,
        desc="Process Items",
        disable=no_bars,
    )

    if not heartbeat_func_kwargs:
        heartbeat_func_kwargs = {}
    if heartbeat_func:
        heartbeat_func(**heartbeat_func_kwargs)

    back_pressure_relief = back_pressured_get.release(processed_items)

    update_items = tqdm(total=total, desc="Update Targets", disable=no_bars)

    async for chunk in grouper(back_pressure_relief, n=builder.chunk_size):
        logger.info(
            f"Processed batch of {builder.chunk_size} items",
            extra={
                "maggma": {
                    "event": "UPDATE",
                    "items": len(chunk),
                    "builder": builder.__class__.__name__,
                    "sources": [source.name for source in builder.sources],
                    "targets": [target.name for target in builder.targets],
                }
            },
        )
        processed_items = [item for item in chunk if item is not None]
        builder.update_targets(processed_items)
        update_items.update(len(processed_items))

    logger.info(
        f"Ended multiprocessing: {builder.__class__.__name__}",
        extra={
            "maggma": {
                "event": "BUILD_ENDED",
                "builder": builder.__class__.__name__,
                "sources": [source.name for source in builder.sources],
                "targets": [target.name for target in builder.targets],
            }
        },
    )

    update_items.close()
    builder.finalize()



"""
Many-to-Many GroupBuilder.
"""

import traceback
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable, Iterator
from datetime import datetime
from math import ceil
from time import time
from typing import Optional

from pydash import get

from maggma.core import Builder, Store
from maggma.utils import Timeout, grouper


class GroupBuilder(Builder, metaclass=ABCMeta):
    """
    Group source docs and produces merged documents for each group
    Supports incremental building, where a source group gets (re)built only if
    it has a newer (by last_updated_field) doc than the corresponding (by key) target doc.

    This is a Many-to-One or Many-to-Many Builder. As a result, this builder can't determine when a source document
    is orphaned.
    """

    def __init__(
        self,
        source: Store,
        target: Store,
        grouping_keys: list[str],
        query: Optional[dict] = None,
        projection: Optional[list] = None,
        timeout: int = 0,
        store_process_time: bool = True,
        retry_failed: bool = False,
        **kwargs,
    ):
        """
        Args:
            source: source store
            target: target store
            query: optional query to filter items from the source store.
            projection: list of keys to project from the source for
                processing. Limits data transfer to improve efficiency.
            delete_orphans: Whether to delete documents on target store
                with key values not present in source store. Deletion happens
                after all updates, during Builder.finalize.
            timeout: maximum running time per item in seconds
            store_process_time: If True, add "_process_time" key to
                document for profiling purposes
            retry_failed: If True, will retry building documents that
                previously failed.
        """
        self.source = source
        self.target = target
        self.grouping_keys = grouping_keys
        self.query = query if query else {}
        self.projection = projection
        self.kwargs = kwargs
        self.timeout = timeout
        self.store_process_time = store_process_time
        self.retry_failed = retry_failed

        self._target_keys_field = f"{self.source.key}s"

        super().__init__(sources=[source], targets=[target], **kwargs)

    def ensure_indexes(self):
        """
        Ensures indices on critical fields for GroupBuilder
        which include the plural version of the target's key field.
        """
        index_checks = [
            self.source.ensure_index(self.source.key),
            self.source.ensure_index(self.source.last_updated_field),
            self.target.ensure_index(self.target.key),
            self.target.ensure_index(self.target.last_updated_field),
            self.target.ensure_index("state"),
            self.target.ensure_index(self._target_keys_field),
        ]

        if not all(index_checks):
            self.logger.warning(
                "Missing one or more important indices on stores. "
                "Performance for large stores may be severely degraded. "
                "Ensure indices on target.key and "
                "[(store.last_updated_field, -1), (store.key, 1)] "
                "for each of source and target."
            )

    def prechunk(self, number_splits: int) -> Iterator[dict]:
        """
        Generic prechunk for group builder to perform domain-decomposition
        by the grouping keys.
        """
        self.ensure_indexes()

        keys = self.get_ids_to_process()
        groups = self.get_groups_from_keys(keys)

        N = ceil(len(groups) / number_splits)
        for split in grouper(keys, N):
            yield {"query": dict(zip(self.grouping_keys, split))}

    def get_items(self):
        self.logger.info(f"Starting {self.__class__.__name__} Builder")

        self.ensure_indexes()
        keys = self.get_ids_to_process()
        groups = self.get_groups_from_keys(keys)

        if self.projection:
            projection = list({*self.projection, self.source.key, self.source.last_updated_field})
        else:
            projection = None

        self.total = len(groups)
        for group in groups:
            group_criteria = dict(zip(self.grouping_keys, group))
            group_criteria.update(self.query)
            yield list(self.source.query(criteria=group_criteria, properties=projection))

    def process_item(self, item: list[dict]) -> dict[tuple, dict]:  # type: ignore
        keys = [d[self.source.key] for d in item]

        self.logger.debug(f"Processing: {keys}")

        time_start = time()

        try:
            with Timeout(seconds=self.timeout):
                processed = self.unary_function(item)
                processed.update({"state": "successful"})
        except Exception as e:
            self.logger.error(traceback.format_exc())
            processed = {"error": str(e), "state": "failed"}

        time_end = time()

        last_updated = [self.source._lu_func[0](d[self.source.last_updated_field]) for d in item]

        update_doc = {
            self.target.key: keys[0],
            f"{self.source.key}s": keys,
            self.target.last_updated_field: max(last_updated),
            "_bt": datetime.utcnow(),
        }
        processed.update({k: v for k, v in update_doc.items() if k not in processed})

        if self.store_process_time:
            processed["_process_time"] = time_end - time_start

        return processed

    def update_targets(self, items: list[dict]):
        """
        Generic update targets for Group Builder.
        """
        target = self.target
        for item in items:
            if "_id" in item:
                del item["_id"]

        if len(items) > 0:
            target.update(items)

    @abstractmethod
    def unary_function(self, items: list[dict]) -> dict:
        """
        Processing function for GroupBuilder.

        Arguments:
            items: list of of documents with matching grouping keys

        Returns:
            Dictionary mapping:
                tuple of source document keys that are in the grouped document
                to the grouped and processed document
        """

    def get_ids_to_process(self) -> Iterable:
        """
        Gets the IDs that need to be processed.
        """
        distinct_from_target = list(self.target.distinct(self._target_keys_field, criteria=self.query))
        processed_ids = []
        # Not always guaranteed that MongoDB will unpack the list so we
        # have to make sure we do that
        for d in distinct_from_target:
            if isinstance(d, list):
                processed_ids.extend(d)
            else:
                processed_ids.append(d)

        all_ids = set(self.source.distinct(self.source.key, criteria=self.query))
        self.logger.debug(f"Found {len(all_ids)} total docs in source")

        if self.retry_failed:
            failed_keys = self.target.distinct(self._target_keys_field, criteria={"state": "failed", **self.query})
            unprocessed_ids = all_ids - (set(processed_ids) - set(failed_keys))
            self.logger.debug(f"Found {len(failed_keys)} failed IDs in target")
        else:
            unprocessed_ids = all_ids - set(processed_ids)

        self.logger.info(f"Found {len(unprocessed_ids)} IDs to process")

        new_ids = set(self.source.newer_in(self.target, criteria=self.query, exhaustive=False))

        self.logger.info(f"Found {len(new_ids)} updated IDs to process")
        return list(new_ids | unprocessed_ids)

    def get_groups_from_keys(self, keys) -> set[tuple]:
        """
        Get the groups by grouping_keys for these documents.
        """
        grouping_keys = self.grouping_keys

        groups: set[tuple] = set()

        for chunked_keys in grouper(keys, self.chunk_size):
            docs = list(
                self.source.query(
                    criteria={self.source.key: {"$in": chunked_keys}},
                    properties=grouping_keys,
                )
            )

            sub_groups = {tuple(get(d, prop, None) for prop in grouping_keys) for d in docs}
            self.logger.debug(f"Found {len(sub_groups)} subgroups to process")

            groups |= sub_groups

        self.logger.info(f"Found {len(groups)} groups to process")
        return groups



# isort: off
from maggma.api.resource.core import HeaderProcessor, HintScheme, Resource

# isort: on

from maggma.api.resource.aggregation import AggregationResource
from maggma.api.resource.post_resource import PostOnlyResource
from maggma.api.resource.read_resource import ReadOnlyResource, attach_query_ops
from maggma.api.resource.s3_url import S3URLResource
from maggma.api.resource.submission import SubmissionResource

__all__ = [
    "Resource",
    "HintScheme",
    "HeaderProcessor",
    "AggregationResource",
    "PostOnlyResource",
    "ReadOnlyResource",
    "attach_query_ops",
    "SubmissionResource",
    "S3URLResource",
]



from maggma.api.query_operator.core import QueryOperator
from maggma.api.query_operator.dynamic import NumericQuery, StringQueryOperator
from maggma.api.query_operator.pagination import PaginationQuery
from maggma.api.query_operator.sorting import SortQuery
from maggma.api.query_operator.sparse_fields import SparseFieldsQuery
from maggma.api.query_operator.submission import SubmissionQuery

__all__ = [
    "QueryOperator",
    "NumericQuery",
    "StringQueryOperator",
    "PaginationQuery",
    "SortQuery",
    "SparseFieldsQuery",
    "SubmissionQuery",
]



import json
from enum import Enum
from random import choice, randint
from typing import Any
from urllib.parse import urlencode

import pytest
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel, Field
from requests import Response
from starlette.testclient import TestClient

from maggma.api.API import API
from maggma.api.query_operator import NumericQuery, PaginationQuery, SparseFieldsQuery, StringQueryOperator
from maggma.api.resource import ReadOnlyResource
from maggma.stores import MemoryStore


class PetType(str, Enum):
    cat = "cat"
    dog = "dog"


class Owner(BaseModel):
    name: str = Field(..., title="Owner's name")
    age: int = Field(..., title="Owne'r Age")
    weight: int = Field(..., title="Owner's weight")


class Pet(BaseModel):
    name: str = Field(..., title="Pet's Name")
    pet_type: PetType = Field(..., title="Pet Type")
    owner_name: str = Field(..., title="Owner's name")


owners = [Owner(name=f"Person{i}", age=randint(10, 100), weight=randint(100, 200)) for i in list(range(10))]


pets = [
    Pet(
        name=f"Pet{i}",
        pet_type=choice(list(PetType)),
        owner_name=choice(owners).name,
    )
    for i in list(range(40))
]


@pytest.fixture()
def owner_store():
    store = MemoryStore("owners", key="name")
    store.connect()
    store.update([jsonable_encoder(d) for d in owners])
    return store


@pytest.fixture()
def pet_store():
    store = MemoryStore("pets", key="name")
    store.connect()
    store.update([jsonable_encoder(d) for d in pets])
    return store


def test_msonable(owner_store, pet_store):
    owner_endpoint = ReadOnlyResource(owner_store, Owner)
    pet_endpoint = ReadOnlyResource(pet_store, Pet)

    manager = API({"owners": owner_endpoint, "pets": pet_endpoint})

    api_dict = manager.as_dict()

    for k in ["@class", "@module", "resources"]:
        assert k in api_dict


def search_helper(payload, base: str = "/?", debug=True) -> tuple[Response, Any]:
    """
    Helper function to directly query search endpoints

    Args:
        store: store f
        base: base of the query, default to /query?
        client: TestClient generated from FastAPI
        payload: query in dictionary format
        debug: True = print out the url, false don't print anything

    Returns:
        request.Response object that contains the response of the corresponding payload
    """
    owner_store = MemoryStore("owners", key="name")
    owner_store.connect()
    owner_store.update([d.model_dump() for d in owners])

    pets_store = MemoryStore("pets", key="name")
    pets_store.connect()
    pets_store.update([jsonable_encoder(d) for d in pets])

    resources = {
        "owners": [
            ReadOnlyResource(
                owner_store,
                Owner,
                query_operators=[
                    StringQueryOperator(model=Owner),  # type: ignore
                    NumericQuery(model=Owner),  # type: ignore
                    SparseFieldsQuery(model=Owner),
                    PaginationQuery(),
                ],
            )
        ],
        "pets": [
            ReadOnlyResource(
                pets_store,
                Owner,
                query_operators=[
                    StringQueryOperator(model=Pet),
                    NumericQuery(model=Pet),
                    SparseFieldsQuery(model=Pet),
                    PaginationQuery(),
                ],
            )
        ],
    }
    api = API(resources=resources)

    client = TestClient(api.app)

    url = base + urlencode(payload)
    if debug:
        print(url)
    res = client.get(url)
    try:
        data = res.json().get("data", [])
    except json.decoder.JSONDecodeError:
        data = res.text

    return res, data


def test_cluster_run(owner_store, pet_store):
    res, data = search_helper(payload="")
    assert res.status_code == 200

    payload = {"name": "Person1", "_limit": 10, "_all_fields": True}
    res, data = search_helper(payload=payload, base="/owners/?")
    assert res.status_code == 200
    assert len(data) == 1
    assert data[0]["name"] == "Person1"

    payload = {"name": "Pet1", "_limit": 10, "_all_fields": True}
    res, data = search_helper(payload=payload, base="/pets/?")
    assert res.status_code == 200
    assert len(data) == 1
    assert data[0]["name"] == "Pet1"



from inspect import signature
from typing import Any, Optional, Union

import orjson
from fastapi import Depends, HTTPException, Path, Request, Response
from pydantic import BaseModel
from pymongo import timeout as query_timeout
from pymongo.errors import NetworkTimeout, PyMongoError

from maggma.api.models import Meta
from maggma.api.models import Response as ResponseModel
from maggma.api.query_operator import PaginationQuery, QueryOperator, SparseFieldsQuery
from maggma.api.resource import HeaderProcessor, HintScheme, Resource
from maggma.api.resource.utils import attach_query_ops, generate_query_pipeline
from maggma.api.utils import STORE_PARAMS, merge_queries, serialization_helper
from maggma.core import Store
from maggma.stores import MongoStore, S3Store


class ReadOnlyResource(Resource):
    """
    Implements a REST Compatible Resource as a GET URL endpoint
    This class provides a number of convenience features
    including full pagination, field projection.
    """

    def __init__(
        self,
        store: Store,
        model: type[BaseModel],
        tags: Optional[list[str]] = None,
        query_operators: Optional[list[QueryOperator]] = None,
        key_fields: Optional[list[str]] = None,
        hint_scheme: Optional[HintScheme] = None,
        header_processor: Optional[HeaderProcessor] = None,
        query_to_configure_on_request: Optional[QueryOperator] = None,
        timeout: Optional[int] = None,
        enable_get_by_key: bool = False,
        enable_default_search: bool = True,
        disable_validation: bool = False,
        query_disk_use: bool = False,
        include_in_schema: Optional[bool] = True,
        sub_path: Optional[str] = "/",
    ):
        """
        Args:
            store: The Maggma Store to get data from
            model: The pydantic model this Resource represents
            tags: List of tags for the Endpoint
            query_operators: Operators for the query language
            hint_scheme: The hint scheme to use for this resource
            header_processor: The header processor to use for this resource
            query_to_configure_on_request: Query operator to configure on request
            timeout: Time in seconds Pymongo should wait when querying MongoDB
                before raising a timeout error
            key_fields: List of fields to always project. Default uses SparseFieldsQuery
                to allow user to define these on-the-fly.
            enable_get_by_key: Enable get by key route for endpoint.
            enable_default_search: Enable default endpoint search behavior.
            query_disk_use: Whether to use temporary disk space in large MongoDB queries.
            disable_validation: Whether to use ORJSON and provide a direct FastAPI response.
                Note this will disable auto JSON serialization and response validation with the
                provided model.
            include_in_schema: Whether the endpoint should be shown in the documented schema.
            sub_path: sub-URL path for the resource.
        """
        self.store = store
        self.tags = tags or []
        self.hint_scheme = hint_scheme
        self.header_processor = header_processor
        self.query_to_configure_on_request = query_to_configure_on_request
        self.key_fields = key_fields
        self.versioned = False
        self.enable_get_by_key = enable_get_by_key
        self.enable_default_search = enable_default_search
        self.timeout = timeout
        self.disable_validation = disable_validation
        self.include_in_schema = include_in_schema
        self.sub_path = sub_path
        self.query_disk_use = query_disk_use

        self.response_model = ResponseModel[model]  # type: ignore

        if not isinstance(store, MongoStore) and self.hint_scheme is not None:
            raise ValueError("Hint scheme is only supported for MongoDB stores")

        self.query_operators = (
            query_operators
            if query_operators is not None
            else [
                PaginationQuery(),
                SparseFieldsQuery(
                    model,
                    default_fields=[self.store.key, self.store.last_updated_field],
                ),
            ]
        )

        super().__init__(model)

    def prepare_endpoint(self):
        """
        Internal method to prepare the endpoint by setting up default handlers
        for routes.
        """
        if self.enable_get_by_key:
            self.build_get_by_key()

        if self.enable_default_search:
            self.build_dynamic_model_search()

    def build_get_by_key(self):
        key_name = self.store.key
        model_name = self.model.__name__

        if self.key_fields is None:
            field_input = SparseFieldsQuery(self.model, [self.store.key, self.store.last_updated_field]).query
        else:

            def field_input():
                return {"properties": self.key_fields}

        def get_by_key(
            request: Request,
            temp_response: Response,
            key: str = Path(
                ...,
                alias=key_name,
                title=f"The {key_name} of the {model_name} to get",
            ),
            _fields: STORE_PARAMS = Depends(field_input),
        ):
            f"""
            Gets a document by the primary key in the store

            Args:
                {key_name}: the id of a single {model_name}

            Returns:
                a single {model_name} document
            """
            self.store.connect()

            try:
                with query_timeout(self.timeout):
                    item = [
                        self.store.query_one(
                            criteria={self.store.key: key},
                            properties=_fields["properties"],
                        )
                    ]
            except (NetworkTimeout, PyMongoError) as e:
                if e.timeout:
                    raise HTTPException(
                        status_code=504,
                        detail="Server timed out trying to obtain data. Try again with a smaller request.",
                    )
                else:
                    raise HTTPException(
                        status_code=500,
                    )

            if item == [None]:
                raise HTTPException(
                    status_code=404,
                    detail=f"Item with {self.store.key} = {key} not found",
                )

            for operator in self.query_operators:
                item = operator.post_process(item, {})

            response = {"data": item}  # type: ignore

            if self.disable_validation:
                response = Response(orjson.dumps(response, default=serialization_helper))  # type: ignore

            if self.header_processor is not None:
                if self.disable_validation:
                    self.header_processor.process_header(response, request)
                else:
                    self.header_processor.process_header(temp_response, request)

            return response

        self.router.get(
            f"{self.sub_path}{{{key_name}}}/",
            summary=f"Get a {model_name} document by by {key_name}",
            response_description=f"Get a {model_name} document by {key_name}",
            response_model=self.response_model,
            response_model_exclude_unset=True,
            tags=self.tags,
            include_in_schema=self.include_in_schema,
        )(get_by_key)

    def build_dynamic_model_search(self):
        model_name = self.model.__name__

        def search(**queries: dict[str, STORE_PARAMS]) -> Union[dict, Response]:
            request: Request = queries.pop("request")  # type: ignore
            temp_response: Response = queries.pop("temp_response")  # type: ignore

            if self.query_to_configure_on_request is not None:
                # give the key name "request", arbitrary choice, as only the value gets merged into the query
                queries["groups"] = self.header_processor.configure_query_on_request(
                    request=request, query_operator=self.query_to_configure_on_request
                )
            # allowed query parameters
            query_params = [
                entry for _, i in enumerate(self.query_operators) for entry in signature(i.query).parameters
            ]
            # check for overlap between allowed query parameters and request query parameters
            overlap = [key for key in request.query_params if key not in query_params]
            if any(overlap):
                if "limit" in overlap or "skip" in overlap:
                    raise HTTPException(
                        status_code=400,
                        detail="'limit' and 'skip' parameters have been renamed. "
                        "Please update your API client to the newest version.",
                    )
                else:
                    raise HTTPException(
                        status_code=400,
                        detail="Request contains query parameters which cannot be used: {}".format(", ".join(overlap)),
                    )

            query: dict[Any, Any] = merge_queries(list(queries.values()))  # type: ignore

            if self.hint_scheme is not None:  # pragma: no cover
                hints = self.hint_scheme.generate_hints(query)
                query.update(hints)

            self.store.connect()

            try:
                with query_timeout(self.timeout):
                    if isinstance(self.store, S3Store):
                        count = self.store.count(criteria=query.get("criteria"))  # type: ignore

                        if self.query_disk_use:
                            data = list(self.store.query(**query, allow_disk_use=True))  # type: ignore
                        else:
                            data = list(self.store.query(**query))
                    else:
                        count = self.store.count(
                            criteria=query.get("criteria"), hint=query.get("count_hint")
                        )  # type: ignore

                        pipeline = generate_query_pipeline(query, self.store)

                        agg_kwargs = {}

                        if query.get("agg_hint"):
                            agg_kwargs["hint"] = query["agg_hint"]

                        data = list(self.store._collection.aggregate(pipeline, **agg_kwargs))

            except (NetworkTimeout, PyMongoError) as e:
                if e.timeout:
                    raise HTTPException(
                        status_code=504,
                        detail="Server timed out trying to obtain data. Try again with a smaller request.",
                    )
                else:
                    raise HTTPException(
                        status_code=500,
                        detail="Server timed out trying to obtain data. Try again with a smaller request,"
                        " or remove sorting fields and sort data locally.",
                    )

            operator_meta = {}

            for operator in self.query_operators:
                data = operator.post_process(data, query)
                operator_meta.update(operator.meta())

            meta = Meta(total_doc=count)

            response = {"data": data, "meta": {**meta.dict(), **operator_meta}}  # type: ignore

            if self.disable_validation:
                response = Response(orjson.dumps(response, default=serialization_helper))  # type: ignore

            if self.header_processor is not None:
                if self.disable_validation:
                    self.header_processor.process_header(response, request)
                else:
                    self.header_processor.process_header(temp_response, request)

            return response

        self.router.get(
            self.sub_path,
            tags=self.tags,
            summary=f"Get {model_name} documents",
            response_model=self.response_model,
            response_description=f"Search for a {model_name}",
            response_model_exclude_unset=True,
        )(attach_query_ops(search, self.query_operators))



"""
Module containing various definitions of Stores.
Stores are a default access pattern to data and provide
various utilities.
"""

import warnings
from collections.abc import Iterator
from itertools import chain, groupby
from pathlib import Path
from typing import Any, Callable, Literal, Optional, Union

import bson
import mongomock
import orjson
from monty.dev import requires
from monty.io import zopen
from monty.json import jsanitize
from monty.serialization import loadfn
from pydash import get, has, set_
from pymongo import MongoClient, ReplaceOne, uri_parser
from pymongo.errors import ConfigurationError, DocumentTooLarge, OperationFailure
from ruamel.yaml import YAML

from maggma.core import Sort, Store, StoreError
from maggma.stores.ssh_tunnel import SSHTunnel
from maggma.utils import confirm_field_index, to_dt

try:
    from montydb import MontyClient, set_storage  # type: ignore
except ImportError:
    MontyClient = None


class MongoStore(Store):
    """
    A Store that connects to a Mongo collection.
    """

    def __init__(
        self,
        database: str,
        collection_name: str,
        host: str = "localhost",
        port: int = 27017,
        username: str = "",
        password: str = "",
        ssh_tunnel: Optional[SSHTunnel] = None,
        safe_update: bool = False,
        auth_source: Optional[str] = None,
        mongoclient_kwargs: Optional[dict] = None,
        default_sort: Optional[dict[str, Union[Sort, int]]] = None,
        **kwargs,
    ):
        """
        Args:
            database: The database name
            collection_name: The collection name
            host: Hostname for the database
            port: TCP port to connect to
            username: Username for the collection
            password: Password to connect with
            safe_update: fail gracefully on DocumentTooLarge errors on update
            auth_source: The database to authenticate on. Defaults to the database name.
            default_sort: Default sort field and direction to use when querying. Can be used to
                ensure determinacy in query results.
        """
        self.database = database
        self.collection_name = collection_name
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.ssh_tunnel = ssh_tunnel
        self.safe_update = safe_update
        self.default_sort = default_sort
        self._coll = None  # type: ignore
        self.kwargs = kwargs

        if auth_source is None:
            auth_source = self.database
        self.auth_source = auth_source
        self.mongoclient_kwargs = mongoclient_kwargs or {}

        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        """
        Return a string representing this data source.
        """
        return f"mongo://{self.host}/{self.database}/{self.collection_name}"

    def connect(self, force_reset: bool = False):
        """
        Connect to the source data.

        Args:
            force_reset: whether to reset the connection or not when the Store is
                already connected.
        """
        if self._coll is None or force_reset:
            if self.ssh_tunnel is None:
                host = self.host
                port = self.port
            else:
                self.ssh_tunnel.start()
                host, port = self.ssh_tunnel.local_address

            conn: MongoClient = (
                MongoClient(
                    host=host,
                    port=port,
                    username=self.username,
                    password=self.password,
                    authSource=self.auth_source,
                    **self.mongoclient_kwargs,
                )
                if self.username != ""
                else MongoClient(host, port, **self.mongoclient_kwargs)
            )
            db = conn[self.database]
            self._coll = db[self.collection_name]  # type: ignore

    def __hash__(self) -> int:
        """Hash for MongoStore."""
        return hash((self.database, self.collection_name, self.last_updated_field))

    @classmethod
    def from_db_file(cls, filename: str, **kwargs):
        """
        Convenience method to construct MongoStore from db_file
        from old QueryEngine format.
        """
        kwargs = loadfn(filename)
        if "collection" in kwargs:
            kwargs["collection_name"] = kwargs.pop("collection")
        # Get rid of aliases from traditional query engine db docs
        kwargs.pop("aliases", None)
        return cls(**kwargs)

    @classmethod
    def from_launchpad_file(cls, lp_file, collection_name, **kwargs):
        """
        Convenience method to construct MongoStore from a launchpad file.

        Note: A launchpad file is a special formatted yaml file used in fireworks

        Returns:
        """
        with open(lp_file) as f:
            yaml = YAML(typ="safe", pure=True)
            lp_creds = yaml.load(f.read())

        db_creds = lp_creds.copy()
        db_creds["database"] = db_creds["name"]
        for key in list(db_creds.keys()):
            if key not in ["database", "host", "port", "username", "password"]:
                db_creds.pop(key)
        db_creds["collection_name"] = collection_name

        return cls(**db_creds, **kwargs)

    def distinct(self, field: str, criteria: Optional[dict] = None, all_exist: bool = False) -> list:
        """
        Get all distinct values for a field.

        Args:
            field: the field(s) to get distinct values for
            criteria: PyMongo filter for documents to search in
        """
        criteria = criteria or {}
        try:
            distinct_vals = self._collection.distinct(field, criteria)
        except (OperationFailure, DocumentTooLarge):
            distinct_vals = [
                d["_id"] for d in self._collection.aggregate([{"$match": criteria}, {"$group": {"_id": f"${field}"}}])
            ]
            if all(isinstance(d, list) for d in filter(None, distinct_vals)):  # type: ignore
                distinct_vals = list(chain.from_iterable(filter(None, distinct_vals)))

        return distinct_vals if distinct_vals is not None else []

    def groupby(
        self,
        keys: Union[list[str], str],
        criteria: Optional[dict] = None,
        properties: Union[dict, list, None] = None,
        sort: Optional[dict[str, Union[Sort, int]]] = None,
        skip: int = 0,
        limit: int = 0,
    ) -> Iterator[tuple[dict, list[dict]]]:
        """
        Simple grouping function that will group documents
        by keys.

        Args:
            keys: fields to group documents
            criteria: PyMongo filter for documents to search in
            properties: properties to return in grouped documents
            sort: Dictionary of sort order for fields. Keys are field names and
                values are 1 for ascending or -1 for descending.
            skip: number documents to skip
            limit: limit on total number of documents returned

        Returns:
            generator returning tuples of (key, list of docs)
        """
        pipeline = []
        if isinstance(keys, str):
            keys = [keys]

        if properties is None:
            properties = []
        if isinstance(properties, dict):
            properties = list(properties.keys())

        if criteria is not None:
            pipeline.append({"$match": criteria})

        if len(properties) > 0:
            pipeline.append({"$project": {p: 1 for p in properties + keys}})

        alpha = "abcdefghijklmnopqrstuvwxyz"
        group_id = {letter: f"${key}" for letter, key in zip(alpha, keys)}
        pipeline.append({"$group": {"_id": group_id, "docs": {"$push": "$$ROOT"}}})
        for d in self._collection.aggregate(pipeline, allowDiskUse=True):
            id_doc = {}  # type: ignore
            for letter, key in group_id.items():
                if has(d["_id"], letter):
                    set_(id_doc, key[1:], d["_id"][letter])
            yield (id_doc, d["docs"])

    @classmethod
    def from_collection(cls, collection):
        """
        Generates a MongoStore from a pymongo collection object
        This is not a fully safe operation as it gives dummy information to the MongoStore
        As a result, this will not serialize and can not reset its connection.

        Args:
            collection: the PyMongo collection to create a MongoStore around
        """
        # TODO: How do we make this safer?
        coll_name = collection.name
        db_name = collection.database.name

        store = cls(db_name, coll_name)
        store._coll = collection
        return store

    @property
    def _collection(self):
        """Property referring to underlying pymongo collection."""
        if self._coll is None:
            raise StoreError("Must connect Mongo-like store before attempting to use it")
        return self._coll

    def count(
        self,
        criteria: Optional[dict] = None,
        hint: Optional[dict[str, Union[Sort, int]]] = None,
    ) -> int:
        """
        Counts the number of documents matching the query criteria.

        Args:
            criteria: PyMongo filter for documents to count in
            hint: Dictionary of indexes to use as hints for query optimizer.
                Keys are field names and values are 1 for ascending or -1 for descending.
        """
        criteria = criteria if criteria else {}

        hint_list = (
            [(k, Sort(v).value) if isinstance(v, int) else (k, v.value) for k, v in hint.items()] if hint else None
        )

        if hint_list is not None:  # pragma: no cover
            return self._collection.count_documents(filter=criteria, hint=hint_list)

        return (
            self._collection.count_documents(filter=criteria)
            if criteria
            else self._collection.estimated_document_count()
        )

    def query(  # type: ignore
        self,
        criteria: Optional[dict] = None,
        properties: Union[dict, list, None] = None,
        sort: Optional[dict[str, Union[Sort, int]]] = None,
        hint: Optional[dict[str, Union[Sort, int]]] = None,
        skip: int = 0,
        limit: int = 0,
        **kwargs,
    ) -> Iterator[dict]:
        """
        Queries the Store for a set of documents.

        Args:
            criteria: PyMongo filter for documents to search in
            properties: properties to return in grouped documents
            sort: Dictionary of sort order for fields. Keys are field names and
                values are 1 for ascending or -1 for descending.
            hint: Dictionary of indexes to use as hints for query optimizer.
                Keys are field names and values are 1 for ascending or -1 for descending.
            skip: number documents to skip
            limit: limit on total number of documents returned
            mongoclient_kwargs: Dict of extra kwargs to pass to pymongo find.
        """
        if isinstance(properties, list):
            properties = {p: 1 for p in properties}

        default_sort_formatted = None

        if self.default_sort is not None:
            default_sort_formatted = [
                (k, Sort(v).value) if isinstance(v, int) else (k, v.value) for k, v in self.default_sort.items()
            ]

        sort_list = (
            [(k, Sort(v).value) if isinstance(v, int) else (k, v.value) for k, v in sort.items()]
            if sort
            else default_sort_formatted
        )

        hint_list = (
            [(k, Sort(v).value) if isinstance(v, int) else (k, v.value) for k, v in hint.items()] if hint else None
        )

        yield from self._collection.find(
            filter=criteria,
            projection=properties,
            skip=skip,
            limit=limit,
            sort=sort_list,
            hint=hint_list,
            **kwargs,
        )

    def ensure_index(self, key: str, unique: Optional[bool] = False) -> bool:
        """
        Tries to create an index and return true if it succeeded.

        Args:
            key: single key to index
            unique: Whether or not this index contains only unique keys.

        Returns:
            bool indicating if the index exists/was created
        """
        if confirm_field_index(self._collection, key):
            return True

        try:
            self._collection.create_index(key, unique=unique, background=True)
            return True
        except Exception:
            return False

    def update(self, docs: Union[list[dict], dict], key: Union[list, str, None] = None):
        """
        Update documents into the Store.

        Args:
            docs: the document or list of documents to update
            key: field name(s) to determine uniqueness for a
                 document, can be a list of multiple fields,
                 a single field, or None if the Store's key
                 field is to be used
        """
        requests = []

        if not isinstance(docs, list):
            docs = [docs]

        for d in (jsanitize(x, allow_bson=True, recursive_msonable=True) for x in docs):
            # document-level validation is optional
            validates = True
            if self.validator:
                validates = self.validator.is_valid(d)
                if not validates:
                    if self.validator.strict:
                        raise ValueError(self.validator.validation_errors(d))
                    self.logger.error(self.validator.validation_errors(d))

            if validates:
                key = key or self.key
                search_doc = {k: d[k] for k in key} if isinstance(key, list) else {key: d[key]}

                requests.append(ReplaceOne(search_doc, d, upsert=True))

        if len(requests) > 0:
            try:
                self._collection.bulk_write(requests, ordered=False)
            except (OperationFailure, DocumentTooLarge) as e:
                if self.safe_update:
                    for req in requests:
                        try:
                            self._collection.bulk_write([req], ordered=False)
                        except (OperationFailure, DocumentTooLarge):
                            self.logger.error(
                                f"Could not upload document for {req._filter} as it was too large for Mongo"
                            )
                else:
                    raise e

    def remove_docs(self, criteria: dict):
        """
        Remove docs matching the query dictionary.

        Args:
            criteria: query dictionary to match
        """
        self._collection.delete_many(filter=criteria)

    def close(self):
        """Close up all collections."""
        self._collection.database.client.close()
        self._coll = None
        if self.ssh_tunnel is not None:
            self.ssh_tunnel.stop()

    def __eq__(self, other: object) -> bool:
        """
        Check equality for MongoStore
        other: other mongostore to compare with.
        """
        if not isinstance(other, MongoStore):
            return False

        fields = ["database", "collection_name", "host", "port", "last_updated_field"]
        return all(getattr(self, f) == getattr(other, f) for f in fields)


class MongoURIStore(MongoStore):
    """
    A Store that connects to a Mongo collection via a URI
    This is expected to be a special mongodb+srv:// URIs that include
    client parameters via TXT records.
    """

    def __init__(
        self,
        uri: str,
        collection_name: str,
        database: Optional[str] = None,
        ssh_tunnel: Optional[SSHTunnel] = None,
        safe_update: bool = False,
        mongoclient_kwargs: Optional[dict] = None,
        default_sort: Optional[dict[str, Union[Sort, int]]] = None,
        **kwargs,
    ):
        """
        Args:
            uri: MongoDB+SRV URI
            database: database to connect to
            collection_name: The collection name
            default_sort: Default sort field and direction to use when querying. Can be used to
                ensure determinacy in query results.
        """
        self.uri = uri
        self.ssh_tunnel = ssh_tunnel
        self.default_sort = default_sort
        self.safe_update = safe_update
        self.mongoclient_kwargs = mongoclient_kwargs or {}

        # parse the dbname from the uri
        if database is None:
            d_uri = uri_parser.parse_uri(uri)
            if d_uri["database"] is None:
                raise ConfigurationError("If database name is not supplied, a database must be set in the uri")
            self.database = d_uri["database"]
        else:
            self.database = database

        self.collection_name = collection_name
        self.kwargs = kwargs
        self._coll = None
        super(MongoStore, self).__init__(**kwargs)  # lgtm

    @property
    def name(self) -> str:
        """
        Return a string representing this data source.
        """
        # TODO: This is not very safe since it exposes the username/password info
        return self.uri

    def connect(self, force_reset: bool = False):
        """
        Connect to the source data.

        Args:
            force_reset: whether to reset the connection or not when the Store is
                already connected.
        """
        if self._coll is None or force_reset:  # pragma: no cover
            conn: MongoClient = MongoClient(self.uri, **self.mongoclient_kwargs)
            db = conn[self.database]
            self._coll = db[self.collection_name]  # type: ignore


class MemoryStore(MongoStore):
    """
    An in-memory Store that functions similarly
    to a MongoStore.
    """

    def __init__(self, collection_name: str = "memory_db", **kwargs):
        """
        Initializes the Memory Store.

        Args:
            collection_name: name for the collection in memory.
        """
        self.collection_name = collection_name
        self.default_sort = None
        self._coll = None
        self.kwargs = kwargs
        super(MongoStore, self).__init__(**kwargs)

    def connect(self, force_reset: bool = False):
        """
        Connect to the source data.

        Args:
            force_reset: whether to reset the connection or not when the Store is
                already connected.
        """
        if self._coll is None or force_reset:
            self._coll = mongomock.MongoClient().db[self.name]  # type: ignore

    def close(self):
        """Close up all collections."""
        self._coll.database.client.close()

    @property
    def name(self):
        """Name for the store."""
        return f"mem://{self.collection_name}"

    def __hash__(self):
        """Hash for the store."""
        return hash((self.name, self.last_updated_field))

    def groupby(
        self,
        keys: Union[list[str], str],
        criteria: Optional[dict] = None,
        properties: Union[dict, list, None] = None,
        sort: Optional[dict[str, Union[Sort, int]]] = None,
        skip: int = 0,
        limit: int = 0,
    ) -> Iterator[tuple[dict, list[dict]]]:
        """
        Simple grouping function that will group documents
        by keys.

        Args:
            keys: fields to group documents
            criteria: PyMongo filter for documents to search in
            properties: properties to return in grouped documents
            sort: Dictionary of sort order for fields. Keys are field names and
                values are 1 for ascending or -1 for descending.
            skip: number documents to skip
            limit: limit on total number of documents returned

        Returns:
            generator returning tuples of (key, list of elements)
        """
        keys = keys if isinstance(keys, list) else [keys]

        if properties is None:
            properties = []
        if isinstance(properties, dict):
            properties = list(properties.keys())

        data = [
            doc for doc in self.query(properties=keys + properties, criteria=criteria) if all(has(doc, k) for k in keys)
        ]

        def grouping_keys(doc):
            return tuple(get(doc, k) for k in keys)

        for vals, group in groupby(sorted(data, key=grouping_keys), key=grouping_keys):
            doc = {}  # type: ignore
            for k, v in zip(keys, vals):
                set_(doc, k, v)
            yield doc, list(group)

    def __eq__(self, other: object) -> bool:
        """
        Check equality for MemoryStore
        other: other MemoryStore to compare with.
        """
        if not isinstance(other, MemoryStore):
            return False

        fields = ["collection_name", "last_updated_field"]
        return all(getattr(self, f) == getattr(other, f) for f in fields)


class JSONStore(MemoryStore):
    """
    A Store for access to a single or multiple JSON files.
    """

    def __init__(
        self,
        paths: Union[str, list[str]],
        read_only: bool = True,
        serialization_option: Optional[int] = None,
        serialization_default: Optional[Callable[[Any], Any]] = None,
        encoding: Optional[str] = None,
        **kwargs,
    ):
        """
        Args:
            paths: paths for json files to turn into a Store
            read_only: whether this JSONStore is read only. When read_only=True,
                       the JSONStore can still apply MongoDB-like writable operations
                       (e.g. an update) because it behaves like a MemoryStore,
                       but it will not write those changes to the file. On the other hand,
                       if read_only=False (i.e., it is writeable), the JSON file
                       will be automatically updated every time a write-like operation is
                       performed.

                       Note that when read_only=False, JSONStore only supports a single JSON
                       file. If the file does not exist, it will be automatically created
                       when the JSONStore is initialized.
            serialization_option:
                option that will be passed to the orjson.dump when saving to the json the file.
            serialization_default:
                default that will be passed to the orjson.dump when saving to the json the file.
            encoding: Character encoding of files to be tracked by the store. The default
                (None) follows python's default behavior, which is to determine the character
                encoding from the platform. This should work in the great majority of cases.
                However, if you encounter a UnicodeDecodeError, consider setting the encoding
                explicitly to 'utf8' or another encoding as appropriate.
        """
        paths = paths if isinstance(paths, (list, tuple)) else [paths]
        self.paths = paths
        self.encoding = encoding

        # file_writable overrides read_only for compatibility reasons
        if "file_writable" in kwargs:
            file_writable = kwargs.pop("file_writable")
            warnings.warn(
                "file_writable is deprecated; use read only instead.",
                DeprecationWarning,
            )
            self.read_only = not file_writable
            if self.read_only != read_only:
                warnings.warn(
                    f"Received conflicting keyword arguments file_writable={file_writable}"
                    f" and read_only={read_only}. Setting read_only={file_writable}.",
                    UserWarning,
                )
        else:
            self.read_only = read_only
        self.kwargs = kwargs

        if not self.read_only and len(paths) > 1:
            raise RuntimeError("Cannot instantiate file-writable JSONStore with multiple JSON files.")

        self.default_sort = None
        self.serialization_option = serialization_option
        self.serialization_default = serialization_default

        super().__init__(**kwargs)

    def connect(self, force_reset: bool = False):
        """
        Loads the files into the collection in memory.

        Args:
            force_reset: whether to reset the connection or not. If False (default) and .connect()
            has been called previously, the .json file will not be read in again. This can improve performance
            on systems with slow storage when multiple connect / disconnects are performed.
        """
        if self._coll is None or force_reset:
            self._coll = mongomock.MongoClient().db[self.name]  # type: ignore

            # create the .json file if it does not exist
            if not self.read_only and not Path(self.paths[0]).exists():
                with zopen(self.paths[0], "w", encoding=self.encoding) as f:
                    data: list[dict] = []
                    bytesdata = orjson.dumps(data)
                    f.write(bytesdata.decode("utf-8"))

            for path in self.paths:
                objects = self.read_json_file(path)
                try:
                    self.update(objects)
                except KeyError:
                    raise KeyError(
                        f"""
                        Key field '{self.key}' not found in {path.name}. This
                        could mean that this JSONStore was initially created with a different key field.
                        The keys found in the .json file are {list(objects[0].keys())}. Try
                        re-initializing your JSONStore using one of these as the key arguments.
                        """
                    )

    def read_json_file(self, path) -> list:
        """
        Helper method to read the contents of a JSON file and generate
        a list of docs.

        Args:
            path: Path to the JSON file to be read
        """
        with zopen(path, "r") as f:
            data = f.read()
            data = data.decode() if isinstance(data, bytes) else data
            objects = bson.json_util.loads(data) if "$oid" in data else orjson.loads(data)
            objects = [objects] if not isinstance(objects, list) else objects
            # datetime objects deserialize to str. Try to convert the last_updated
            # field back to datetime.
            # # TODO - there may still be problems caused if a JSONStore is init'ed from
            # documents that don't contain a last_updated field
            # See Store.last_updated in store.py.
            for obj in objects:
                if obj.get(self.last_updated_field):
                    obj[self.last_updated_field] = to_dt(obj[self.last_updated_field])

        return objects

    def update(self, docs: Union[list[dict], dict], key: Union[list, str, None] = None):
        """
        Update documents into the Store.

        For a file-writable JSONStore, the json file is updated.

        Args:
            docs: the document or list of documents to update
            key: field name(s) to determine uniqueness for a
                 document, can be a list of multiple fields,
                 a single field, or None if the Store's key
                 field is to be used
        """
        super().update(docs=docs, key=key)
        if not self.read_only:
            self.update_json_file()

    def remove_docs(self, criteria: dict):
        """
        Remove docs matching the query dictionary.

        For a file-writable JSONStore, the json file is updated.

        Args:
            criteria: query dictionary to match
        """
        super().remove_docs(criteria=criteria)
        if not self.read_only:
            self.update_json_file()

    def update_json_file(self):
        """
        Updates the json file when a write-like operation is performed.
        """
        with zopen(self.paths[0], "w", encoding=self.encoding) as f:
            data = list(self.query())
            for d in data:
                d.pop("_id")
            bytesdata = orjson.dumps(
                data,
                option=self.serialization_option,
                default=self.serialization_default,
            )
            f.write(bytesdata.decode("utf-8"))

    def __hash__(self):
        return hash((*self.paths, self.last_updated_field))

    def __eq__(self, other: object) -> bool:
        """
        Check equality for JSONStore.

        Args:
            other: other JSONStore to compare with
        """
        if not isinstance(other, JSONStore):
            return False

        fields = ["paths", "last_updated_field"]
        return all(getattr(self, f) == getattr(other, f) for f in fields)


@requires(
    MontyClient is not None,
    "MontyStore requires MontyDB to be installed. See the MontyDB repository for more "
    "information: https://github.com/davidlatwe/montydb",
)
class MontyStore(MemoryStore):
    """
    A MongoDB compatible store that uses on disk files for storage.

    This is handled under the hood using MontyDB. A number of on-disk storage options
    are available but MontyDB provides a mongo style interface for all options. The
    options include:

    - sqlite: Uses an sqlite database to store documents.
    - lightning: Uses Lightning Memory-Mapped Database (LMDB) for storage. This can
      provide fast read and write times but requires lmdb to be installed (in most cases
      this can be achieved using ``pip install lmdb``).
    - flatfile: Uses a system of flat json files. This is not recommended as multiple
      simultaneous connections to the store will not work correctly.

    Note that MontyDB (and, therefore, MontyStore) will write out a new database to
    the disk but cannot be used to read an existing (e.g. SQLite) database that wasn't
    formatted by MontyDB.

    See the MontyDB repository for more information: https://github.com/davidlatwe/montydb
    """

    def __init__(
        self,
        collection_name,
        database_path: Optional[str] = None,
        database_name: str = "db",
        storage: Literal["sqlite", "flatfile", "lightning"] = "sqlite",
        storage_kwargs: Optional[dict] = None,
        client_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initializes the Monty Store.

        Args:
            collection_name: Name for the collection.
            database_path: Path to on-disk database files. If None, the current working
                directory will be used.
            database_name: The database name.
            storage: The storage type. Options include "sqlite", "lightning", "flatfile". Note that
            although MontyDB supports in memory storage, this capability is disabled in maggma to avoid unintended
            behavior, since multiple in-memory MontyStore would actually point to the same data.
            storage_kwargs: Keyword arguments passed to ``montydb.set_storage``.
            client_kwargs: Keyword arguments passed to the ``montydb.MontyClient``
                constructor.
            **kwargs: Additional keyword arguments passed to the Store constructor.
        """
        if database_path is None:
            database_path = str(Path.cwd())

        self.database_path = database_path
        self.database_name = database_name
        self.collection_name = collection_name
        self._coll = None  # type: ignore
        self.default_sort = None
        self.ssh_tunnel = None  # This is to fix issues with the tunnel on close
        self.kwargs = kwargs
        self.storage = storage
        self.storage_kwargs = storage_kwargs or {
            "use_bson": True,  # import pymongo's BSON; do not use montydb's
            "mongo_version": "4.0",
        }
        self.client_kwargs = client_kwargs or {}
        super(MongoStore, self).__init__(**kwargs)

    def connect(self, force_reset: bool = False):
        """
        Connect to the database store.

        Args:
            force_reset: whether to reset the connection or not when the Store is
                already connected.
        """
        if not self._coll or force_reset:
            # TODO - workaround, may be obviated by a future montydb update
            if self.database_path != ":memory:":
                set_storage(self.database_path, storage=self.storage, **self.storage_kwargs)
            client = MontyClient(self.database_path, **self.client_kwargs)
            self._coll = client[self.database_name][self.collection_name]

    @property
    def name(self) -> str:
        """Return a string representing this data source."""
        return f"monty://{self.database_path}/{self.database_name}/{self.collection_name}"

    def count(
        self,
        criteria: Optional[dict] = None,
        hint: Optional[dict[str, Union[Sort, int]]] = None,
    ) -> int:
        """
        Counts the number of documents matching the query criteria.

        Args:
            criteria: PyMongo filter for documents to count in
            hint: Dictionary of indexes to use as hints for query optimizer.
                Keys are field names and values are 1 for ascending or -1 for descending.
        """
        criteria = criteria if criteria else {}

        hint_list = (
            [(k, Sort(v).value) if isinstance(v, int) else (k, v.value) for k, v in hint.items()] if hint else None
        )

        if hint_list is not None:  # pragma: no cover
            return self._collection.count_documents(filter=criteria, hint=hint_list)

        return self._collection.count_documents(filter=criteria)

    def update(self, docs: Union[list[dict], dict], key: Union[list, str, None] = None):
        """
        Update documents into the Store.

        Args:
            docs: The document or list of documents to update.
            key: Field name(s) to determine uniqueness for a document, can be a list of
                multiple fields, a single field, or None if the Store's key field is to be
                used.
        """
        if not isinstance(docs, list):
            docs = [docs]

        for d in docs:
            d = jsanitize(d, allow_bson=True)

            # document-level validation is optional
            validates = True
            if self.validator:
                validates = self.validator.is_valid(d)
                if not validates:
                    if self.validator.strict:
                        raise ValueError(self.validator.validation_errors(d))
                    self.logger.error(self.validator.validation_errors(d))

            if validates:
                key = key or self.key
                search_doc = {k: d[k] for k in key} if isinstance(key, list) else {key: d[key]}

                self._collection.replace_one(search_doc, d, upsert=True)



import pymongo
import pytest
from pymongo.errors import DocumentTooLarge, OperationFailure

from maggma.stores import GridFSStore, MemoryStore, MongoStore
from maggma.stores.shared_stores import MultiStore, StoreFacade
from maggma.validators import JSONSchemaValidator


@pytest.fixture()
def mongostore():
    store = MongoStore("maggma_test", "test")
    store.connect()
    yield store
    store._collection.drop()


@pytest.fixture()
def gridfsstore():
    store = GridFSStore("maggma_test", "test", key="task_id")
    store.connect()
    yield store
    store._files_collection.drop()
    store._chunks_collection.drop()


@pytest.fixture()
def multistore():
    return MultiStore()


@pytest.fixture()
def memorystore():
    store = MemoryStore()
    store.connect()
    return store


def test_add_stores(multistore, mongostore, gridfsstore):
    # Should be empty at the start
    assert multistore.count_stores() == 0

    multistore.ensure_store(mongostore)
    assert multistore.count_stores() == 1
    assert multistore.get_store_index(mongostore) == 0

    # Attempting to reinsert this store should do nothing
    multistore.ensure_store(mongostore)
    assert multistore.count_stores() == 1

    # Make a copy of the mongostore and it should still do nothing
    temp_mongostore = MongoStore.from_dict(mongostore.as_dict())
    multistore.ensure_store(temp_mongostore)
    assert multistore.count_stores() == 1
    assert multistore.get_store_index(temp_mongostore) == 0
    # Add this copy again, but don't use ensure_store
    # This tests the case in which a prior thread added
    # the store, but this current process was already
    # waiting for the lock acquisition
    multistore.add_store(temp_mongostore)
    assert multistore.count_stores() == 1

    # Add the GridFSStore to the MultiStore()
    multistore.ensure_store(gridfsstore)
    assert multistore.count_stores() == 2
    assert multistore.get_store_index(gridfsstore) == 1

    # Add something that isn't a store
    class DummyObject:
        def __init__(self, a: int):
            self.a = a

    with pytest.raises(TypeError):
        multistore.ensure_store(DummyObject(1))


def test_store_facade(multistore, mongostore, gridfsstore):
    StoreFacade(mongostore, multistore)
    assert multistore.count_stores() == 1
    assert multistore.get_store_index(mongostore) == 0

    StoreFacade(gridfsstore, multistore)
    assert multistore.count_stores() == 2
    assert multistore.get_store_index(gridfsstore) == 1


def test_multistore_query(multistore, mongostore, memorystore):
    memorystore_facade = StoreFacade(memorystore, multistore)
    mongostore_facade = StoreFacade(mongostore, multistore)
    temp_mongostore_facade = StoreFacade(MongoStore.from_dict(mongostore.as_dict()), multistore)

    memorystore_facade._collection.insert_one({"a": 1, "b": 2, "c": 3})
    assert memorystore_facade.query_one(properties=["a"])["a"] == 1
    assert memorystore_facade.query_one(properties=["a"])["a"] == 1
    assert memorystore_facade.query_one(properties=["b"])["b"] == 2
    assert memorystore_facade.query_one(properties=["c"])["c"] == 3

    mongostore_facade._collection.insert_one({"a": 4, "b": 5, "c": 6})
    assert mongostore_facade.query_one(properties=["a"])["a"] == 4
    assert mongostore_facade.query_one(properties=["a"])["a"] == 4
    assert mongostore_facade.query_one(properties=["b"])["b"] == 5
    assert mongostore_facade.query_one(properties=["c"])["c"] == 6

    assert temp_mongostore_facade.query_one(properties=["a"])["a"] == 4
    assert temp_mongostore_facade.query_one(properties=["a"])["a"] == 4
    assert temp_mongostore_facade.query_one(properties=["b"])["b"] == 5
    assert temp_mongostore_facade.query_one(properties=["c"])["c"] == 6


def test_multistore_count(multistore, mongostore, memorystore):
    memorystore_facade = StoreFacade(memorystore, multistore)

    memorystore_facade._collection.insert_one({"a": 1, "b": 2, "c": 3})
    assert memorystore_facade.count() == 1
    memorystore_facade._collection.insert_one({"aa": 1, "b": 2, "c": 3})
    assert memorystore_facade.count() == 2
    assert memorystore_facade.count({"a": 1}) == 1


def test_multistore_distinct(multistore, mongostore):
    mongostore_facade = StoreFacade(mongostore, multistore)

    mongostore_facade._collection.insert_one({"a": 1, "b": 2, "c": 3})
    mongostore_facade._collection.insert_one({"a": 4, "d": 5, "e": 6, "g": {"h": 1}})
    assert set(mongostore_facade.distinct("a")) == {1, 4}

    # Test list distinct functionality
    mongostore_facade._collection.insert_one({"a": 4, "d": 6, "e": 7})
    mongostore_facade._collection.insert_one({"a": 4, "d": 6, "g": {"h": 2}})

    # Test distinct subdocument functionality
    ghs = mongostore_facade.distinct("g.h")
    assert set(ghs) == {1, 2}

    # Test when key doesn't exist
    assert mongostore_facade.distinct("blue") == []

    # Test when null is a value
    mongostore_facade._collection.insert_one({"i": None})
    assert mongostore_facade.distinct("i") == [None]

    # Test to make sure DocumentTooLarge errors get dealt with properly using built in distinct
    mongostore_facade._collection.insert_many([{"key": [f"mp-{i}"]} for i in range(1000000)])
    vals = mongostore_facade.distinct("key")
    # Test to make sure distinct on array field is unraveled when using manual distinct
    assert len(vals) == len(list(range(1000000)))
    assert all(isinstance(v, str) for v in vals)

    # Test to make sure manual distinct uses the criteria query
    mongostore_facade._collection.insert_many([{"key": f"mp-{i}", "a": 2} for i in range(1000001, 2000001)])
    vals = mongostore_facade.distinct("key", {"a": 2})
    assert len(vals) == len(list(range(1000001, 2000001)))


def test_multistore_update(multistore, mongostore):
    mongostore_facade = StoreFacade(mongostore, multistore)

    mongostore_facade.update({"e": 6, "d": 4}, key="e")
    assert mongostore_facade.query_one(criteria={"d": {"$exists": 1}}, properties=["d"])["d"] == 4

    mongostore_facade.update([{"e": 7, "d": 8, "f": 9}], key=["d", "f"])
    assert mongostore_facade.query_one(criteria={"d": 8, "f": 9}, properties=["e"])["e"] == 7

    mongostore_facade.update([{"e": 11, "d": 8, "f": 9}], key=["d", "f"])
    assert mongostore_facade.query_one(criteria={"d": 8, "f": 9}, properties=["e"])["e"] == 11

    test_schema = {
        "type": "object",
        "properties": {"e": {"type": "integer"}},
        "required": ["e"],
    }
    mongostore_facade.validator = JSONSchemaValidator(schema=test_schema)
    mongostore_facade.update({"e": 100, "d": 3}, key="e")

    # Continue to update doc when validator is not set to strict mode
    mongostore_facade.update({"e": "abc", "d": 3}, key="e")

    # ensure safe_update works to not throw DocumentTooLarge errors
    large_doc = {f"mp-{i}": f"mp-{i}" for i in range(1000000)}
    large_doc["e"] = 999
    with pytest.raises((OperationFailure, DocumentTooLarge)):
        mongostore_facade.update([large_doc, {"e": 1001}], key="e")

    mongostore_facade.safe_update = True
    assert mongostore_facade.safe_update is True
    mongostore_facade.update([large_doc, {"e": 1001}], key="e")
    assert mongostore_facade.query_one({"e": 1001}) is not None


def test_multistore_groupby(multistore, mongostore):
    mongostore_facade = StoreFacade(mongostore, multistore)

    mongostore_facade.update(
        [
            {"e": 7, "d": 9, "f": 9},
            {"e": 7, "d": 9, "f": 10},
            {"e": 8, "d": 9, "f": 11},
            {"e": 9, "d": 10, "f": 12},
        ],
        key="f",
    )
    data = list(mongostore_facade.groupby("d"))
    assert len(data) == 2
    grouped_by_9 = next(g[1] for g in data if g[0]["d"] == 9)
    assert len(grouped_by_9) == 3
    grouped_by_10 = next(g[1] for g in data if g[0]["d"] == 10)
    assert len(grouped_by_10) == 1

    data = list(mongostore_facade.groupby(["e", "d"]))
    assert len(data) == 3


def test_multistore_remove_docs(multistore, mongostore):
    mongostore_facade = StoreFacade(mongostore, multistore)

    mongostore_facade._collection.insert_one({"a": 1, "b": 2, "c": 3})
    mongostore_facade._collection.insert_one({"a": 4, "d": 5, "e": 6, "g": {"h": 1}})
    mongostore_facade.remove_docs({"a": 1})
    assert len(list(mongostore_facade.query({"a": 4}))) == 1
    assert len(list(mongostore_facade.query({"a": 1}))) == 0


def test_multistore_connect_reconnect(multistore, mongostore):
    mongostore_facade = StoreFacade(mongostore, multistore)

    assert isinstance(mongostore_facade._collection, pymongo.collection.Collection)
    mongostore_facade.close()
    assert mongostore_facade._coll is None
    mongostore_facade.connect()

    # Test using the multistore to close connections
    multistore.close_all()
    assert mongostore_facade._coll is None
    multistore.connect_all()
    assert isinstance(mongostore_facade._collection, pymongo.collection.Collection)


def test_multistore_name(multistore, mongostore):
    mongostore_facade = StoreFacade(mongostore, multistore)

    assert mongostore_facade.name == "mongo://localhost/maggma_test/test"


def test_multistore_ensure_index(multistore, mongostore):
    mongostore_facade = StoreFacade(mongostore, multistore)
    assert mongostore_facade.ensure_index("test_key")
    # TODO: How to check for exception?



"""Primary Maggma module."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("maggma")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    pass



from datetime import datetime
from enum import Enum
from typing import Union

import pytest
from bson import ObjectId
from monty.json import MSONable
from pydantic import BaseModel, Field

from maggma.api.utils import api_sanitize, serialization_helper


class SomeEnum(Enum):
    A = 1
    B = 2
    C = 3


class Pet(MSONable):
    def __init__(self, name, age):
        self.name = name
        self.age = age


class AnotherPet(MSONable):
    def __init__(self, name, age):
        self.name = name
        self.age = age


class AnotherOwner(BaseModel):
    name: str = Field(..., description="Owner name")
    weight_or_pet: Union[float, AnotherPet] = Field(..., title="Owners weight or Pet")


class Owner(BaseModel):
    name: str = Field(..., title="Owner's name")
    age: int = Field(..., title="Owne'r Age")
    weight: float = Field(..., title="Owner's weight")
    last_updated: datetime = Field(..., title="Last updated date for this record")
    pet: Pet = Field(..., title="Owner's Pet")
    other: SomeEnum = Field(..., title="A enum?")


def test_api_sanitize():
    # Ensure model validation fails
    with pytest.raises(ValueError):
        Owner()

    # This should still fail validation
    new_owner = api_sanitize(Owner, fields_to_leave=["Owner.name"])
    with pytest.raises(ValueError):
        new_owner()

    new_owner(name="owner")

    # These will fail if non-optional fields are not turned off
    new_owner2 = api_sanitize(Owner)
    new_owner()  # api_sanitize is in-place
    new_owner2()
    Owner()

    # This should fail type validation for pet
    with pytest.raises(Exception):
        Owner(pet="fido")

    temp_pet_dict = Pet(name="fido", age=3).as_dict()
    bad_pet_dict = dict(temp_pet_dict)
    del bad_pet_dict["@module"]
    del bad_pet_dict["@class"]

    # This should fail because of bad data type
    with pytest.raises(Exception):
        Owner(pet=bad_pet_dict)

    assert isinstance(Owner(pet=temp_pet_dict).pet, Pet)

    api_sanitize(Owner, allow_dict_msonable=True)

    # This should still fail because of bad data type
    with pytest.raises(Exception):
        Owner(pet=bad_pet_dict)

    # This should work
    assert isinstance(Owner(pet=temp_pet_dict).pet, dict)

    # This should work evne though AnotherPet is inside the Union type
    api_sanitize(AnotherOwner, allow_dict_msonable=True)
    temp_pet_dict = AnotherPet(name="fido", age=3).as_dict()

    assert isinstance(AnotherPet.validate_monty_v2(temp_pet_dict, None), dict)


def test_serialization_helper():
    oid = ObjectId("60b7d47bb671aa7b01a2adf6")
    assert serialization_helper(oid) == "60b7d47bb671aa7b01a2adf6"


@pytest.mark.xfail()
def test_serialization_helper_xfail():
    oid = "test"
    serialization_helper(oid)



from datetime import datetime
from typing import Optional

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from monty.json import MSONable
from starlette.responses import RedirectResponse

from maggma.api.resource import Resource


class API(MSONable):
    """
    Basic API manager to tie together various resources.
    """

    def __init__(
        self,
        resources: dict[str, list[Resource]],
        title: str = "Generic API",
        version: str = "v0.0.0",
        debug: bool = False,
        heartbeat_meta: Optional[dict] = None,
        description: Optional[str] = None,
        tags_meta: Optional[list[dict]] = None,
    ):
        """
        Args:
            resources: dictionary of resource objects and http prefix they live in
            title: a string title for this API
            version: the version for this API
            debug: turns debug on in FastAPI
            heartbeat_meta: dictionary of additional metadata to include in the heartbeat response
            description: description of the API to be used in the generated docs
            tags_meta: descriptions of tags to be used in the generated docs.
        """
        self.title = title
        self.version = version
        self.debug = debug
        self.heartbeat_meta = heartbeat_meta
        self.description = description
        self.tags_meta = tags_meta

        if len(resources) == 0:
            raise RuntimeError("ERROR: There are no endpoints provided")

        self.resources = resources

    def on_startup(self):
        """
        Basic startup that runs the resource startup functions.
        """
        for resource_list in self.resources.values():
            for resource in resource_list:
                resource.on_startup()

    @property
    def app(self):
        """
        App server for the cluster manager.
        """
        app = FastAPI(
            title=self.title,
            version=self.version,
            on_startup=[self.on_startup],
            debug=self.debug,
            description=self.description,
            openapi_tags=self.tags_meta,
        )

        # Allow requests from other domains in debug mode. This allows
        # testing with local deployments of other services. For production
        # deployment, this will be taken care of by nginx.
        if self.debug:
            app.add_middleware(
                CORSMiddleware,
                allow_origins=["*"],
                allow_methods=["GET"],
                allow_headers=["*"],
            )

        for prefix, resource_list in self.resources.items():
            main_resource = resource_list.pop(0)
            for resource in resource_list:
                main_resource.router.include_router(resource.router)

            app.include_router(main_resource.router, prefix=f"/{prefix}")

        app.add_middleware(GZipMiddleware, minimum_size=1000)

        @app.get("/heartbeat", include_in_schema=False)
        @app.head("/heartbeat", include_in_schema=False)
        def heartbeat():
            """API Heartbeat for Load Balancing."""
            return {
                "status": "OK",
                "time": datetime.utcnow(),
                "version": self.version,
                **self.heartbeat_meta,
            }

        @app.get("/", include_in_schema=False)
        def redirect_docs():
            """Redirects the root end point to the docs."""
            return RedirectResponse(url=app.docs_url, status_code=301)

        return app

    def run(self, ip: str = "127.0.0.1", port: int = 8000, log_level: str = "info"):
        """
        Runs the Cluster Manager locally.

        Args:
            ip: Local IP to listen on
            port: Local port to listen on
            log_level: Logging level for the webserver

        Returns:
            None
        """
        uvicorn.run(self.app, host=ip, port=port, log_level=log_level, reload=False)



from typing import Any, Optional

import orjson
from fastapi import HTTPException, Request, Response
from pydantic import BaseModel
from pymongo import timeout as query_timeout
from pymongo.errors import NetworkTimeout, PyMongoError

from maggma.api.models import Meta
from maggma.api.models import Response as ResponseModel
from maggma.api.query_operator import QueryOperator
from maggma.api.resource import HeaderProcessor, Resource
from maggma.api.resource.utils import attach_query_ops
from maggma.api.utils import STORE_PARAMS, merge_queries, serialization_helper
from maggma.core import Store


class AggregationResource(Resource):
    """
    Implements a REST Compatible Resource as a GET URL endpoint.
    """

    def __init__(
        self,
        store: Store,
        model: type[BaseModel],
        pipeline_query_operator: QueryOperator,
        timeout: Optional[int] = None,
        tags: Optional[list[str]] = None,
        include_in_schema: Optional[bool] = True,
        sub_path: Optional[str] = "/",
        header_processor: Optional[HeaderProcessor] = None,
    ):
        """
        Args:
            store: The Maggma Store to get data from
            model: The pydantic model this Resource represents
            tags: List of tags for the Endpoint
            pipeline_query_operator: Operator for the aggregation pipeline
            timeout: Time in seconds Pymongo should wait when querying MongoDB
                before raising a timeout error
            include_in_schema: Whether the endpoint should be shown in the documented schema.
            sub_path: sub-URL path for the resource.
        """
        self.store = store
        self.tags = tags or []

        self.include_in_schema = include_in_schema
        self.sub_path = sub_path
        self.response_model = ResponseModel[model]  # type: ignore

        self.pipeline_query_operator = pipeline_query_operator
        self.header_processor = header_processor
        self.timeout = timeout

        super().__init__(model)

    def prepare_endpoint(self):
        """
        Internal method to prepare the endpoint by setting up default handlers
        for routes.
        """
        self.build_dynamic_model_search()

    def build_dynamic_model_search(self):
        model_name = self.model.__name__

        def search(**queries: dict[str, STORE_PARAMS]) -> dict:
            request: Request = queries.pop("request")  # type: ignore
            queries.pop("temp_response")  # type: ignore

            query: dict[Any, Any] = merge_queries(list(queries.values()))  # type: ignore

            self.store.connect()

            try:
                with query_timeout(self.timeout):
                    data = list(self.store._collection.aggregate(query["pipeline"]))
            except (NetworkTimeout, PyMongoError) as e:
                if e.timeout:
                    raise HTTPException(
                        status_code=504,
                        detail="Server timed out trying to obtain data. Try again with a smaller request.",
                    )
                else:
                    raise HTTPException(
                        status_code=500,
                    )

            count = len(data)

            data = self.pipeline_query_operator.post_process(data, query)
            operator_meta = self.pipeline_query_operator.meta()

            meta = Meta(total_doc=count)
            response = {"data": data, "meta": {**meta.dict(), **operator_meta}}
            response = Response(orjson.dumps(response, default=serialization_helper))  # type: ignore

            if self.header_processor is not None:
                self.header_processor.process_header(response, request)

            return response

        self.router.get(
            self.sub_path,
            tags=self.tags,
            summary=f"Get {model_name} documents",
            response_model=self.response_model,
            response_description=f"Get {model_name} data",
            response_model_exclude_unset=True,
        )(attach_query_ops(search, [self.pipeline_query_operator]))



"""
Azure testing requires Azurite.
It can be set up according to the instructions:
https://github.com/Azure/Azurite
With docker can be started by running:
docker run -p 10000:10000 mcr.microsoft.com/azure-storage/azurite azurite-blob --blobHost 0.0.0.0
"""

import time
from contextlib import contextmanager
from datetime import datetime

import pytest

from maggma.stores import AzureBlobStore, MemoryStore, MongoStore

try:
    import azure.storage.blob as azure_blob
    from azure.storage.blob import BlobServiceClient
except (ImportError, ModuleNotFoundError):
    azure_blob = None  # type: ignore


AZURITE_CONNECTION_STRING = (
    "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;"
    "AccountKey="
    "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq"
    "/K1SZFPTOtr/KBHBeksoGMGw==;"
    "BlobEndpoint=http://127.0.0.1:10000/devstoreaccount1;"
)

AZURITE_CONTAINER_NAME = "maggma-test-container"


@pytest.fixture()
def mongostore():
    store = MongoStore("maggma_test", "test")
    store.connect()
    yield store
    store._collection.drop()


# a context manager and not a fixture to handle multiple containers in the
# same test
@contextmanager
def azurite_container(container_name=AZURITE_CONTAINER_NAME, create_container=True):
    if azure_blob is None:
        pytest.skip("azure-storage-blob is required to test AzureBlobStore")

    blob_service_client = BlobServiceClient.from_connection_string(AZURITE_CONNECTION_STRING)

    container_client = blob_service_client.get_container_client(container_name)
    if container_client.exists():
        container_client.delete_container()

    if create_container:
        container_client.create_container()

    try:
        yield
    finally:
        if container_client.exists():
            container_client.delete_container()


@pytest.fixture()
def blobstore():
    with azurite_container():
        index = MemoryStore("index", key="task_id")
        store = AzureBlobStore(
            index,
            container_name=AZURITE_CONTAINER_NAME,
            azure_client_info={"connection_string": AZURITE_CONNECTION_STRING},
        )
        store.connect()

        yield store


@pytest.fixture()
def blobstore_two_docs(blobstore):
    blobstore.update(
        [
            {
                "task_id": "mp-1",
                "data": "asd",
                blobstore.last_updated_field: datetime.utcnow(),
            }
        ]
    )
    blobstore.update(
        [
            {
                "task_id": "mp-3",
                "data": "sdf",
                blobstore.last_updated_field: datetime.utcnow(),
            }
        ]
    )

    return blobstore


@pytest.fixture()
def blobstore_w_subdir():
    with azurite_container():
        index = MemoryStore("index")
        store = AzureBlobStore(
            index,
            container_name=AZURITE_CONTAINER_NAME,
            azure_client_info={"connection_string": AZURITE_CONNECTION_STRING},
            sub_dir="subdir1",
        )
        store.connect()

        yield store


@pytest.fixture()
def blobstore_multi(blobstore):
    blobstore.workers = 4

    return blobstore


def test_keys():
    with azurite_container():
        index = MemoryStore("index", key=1)
        with pytest.raises(AssertionError, match=r"Since we are.*"):
            store = AzureBlobStore(index, AZURITE_CONTAINER_NAME, workers=4, key=1)
        index = MemoryStore("index", key="key1")
        with pytest.warns(UserWarning, match=r"The desired .*key.*$"):
            store = AzureBlobStore(
                index,
                AZURITE_CONTAINER_NAME,
                workers=4,
                key="key2",
                azure_client_info={"connection_string": AZURITE_CONNECTION_STRING},
            )
        store.connect()
        store.update({"key1": "mp-1", "data": "1234"})
        with pytest.raises(KeyError):
            store.update({"key2": "mp-2", "data": "1234"})
        assert store.key == store.index.key == "key1"


def test_multi_update(blobstore_two_docs, blobstore_multi):
    data = [
        {
            "task_id": str(j),
            "data": "DATA",
            blobstore_multi.last_updated_field: datetime.utcnow(),
        }
        for j in range(32)
    ]

    def fake_writing(doc, search_keys):
        time.sleep(0.20)
        return {k: doc[k] for k in search_keys}

    blobstore_two_docs.write_doc_to_blob = fake_writing
    blobstore_multi.write_doc_to_blob = fake_writing

    start = time.time()
    blobstore_multi.update(data, key=["task_id"])
    end = time.time()
    time_multi = end - start

    start = time.time()
    blobstore_two_docs.update(data, key=["task_id"])
    end = time.time()
    time_single = end - start
    assert time_single > time_multi * (blobstore_multi.workers - 1) / (blobstore_two_docs.workers)


def test_count(blobstore_two_docs):
    assert blobstore_two_docs.count() == 2
    assert blobstore_two_docs.count({"task_id": "mp-3"}) == 1


def test_query(blobstore_two_docs):
    assert blobstore_two_docs.query_one(criteria={"task_id": "mp-2"}) is None
    assert blobstore_two_docs.query_one(criteria={"task_id": "mp-1"})["data"] == "asd"
    assert blobstore_two_docs.query_one(criteria={"task_id": "mp-3"})["data"] == "sdf"

    assert len(list(blobstore_two_docs.query())) == 2


def test_update(blobstore_two_docs):
    blobstore_two_docs.update(
        [
            {
                "task_id": "mp-199999",
                "data": "asd",
                blobstore_two_docs.last_updated_field: datetime.utcnow(),
            }
        ]
    )
    assert blobstore_two_docs.query_one({"task_id": "mp-199999"}) is not None

    blobstore_two_docs.compress = True
    blobstore_two_docs.update([{"task_id": "mp-4", "data": "asd"}])
    obj = blobstore_two_docs.index.query_one({"task_id": "mp-4"})
    assert obj["compression"] == "zlib"
    assert obj["obj_hash"] == "be74de5ac71f00ec9e96441a3c325b0592c07f4c"
    assert blobstore_two_docs.query_one({"task_id": "mp-4"})["data"] == "asd"


def test_rebuild_meta_from_index(blobstore_two_docs):
    blobstore_two_docs.update([{"task_id": "mp-2", "data": "asd"}])
    blobstore_two_docs.index.update({"task_id": "mp-2", "add_meta": "hello"})
    blobstore_two_docs.rebuild_metadata_from_index()
    print(list(blobstore_two_docs.container.list_blobs()))
    blob_client = blobstore_two_docs.container.get_blob_client("mp-2")
    metadata = blob_client.get_blob_properties()["metadata"]
    assert metadata["add_meta"] == "hello"


def test_rebuild_index(blobstore_two_docs):
    blobstore_two_docs.update([{"task_id": "mp-2", "data": "asd"}])
    assert (
        blobstore_two_docs.index.query_one({"task_id": "mp-2"})["obj_hash"]
        == "a69fe0c2cca3a3384c2b1d2f476972704f179741"
    )
    blobstore_two_docs.index.remove_docs({})
    assert blobstore_two_docs.index.query_one({"task_id": "mp-2"}) is None
    blobstore_two_docs.rebuild_index_from_blob_data()
    assert (
        blobstore_two_docs.index.query_one({"task_id": "mp-2"})["obj_hash"]
        == "a69fe0c2cca3a3384c2b1d2f476972704f179741"
    )


def tests_msonable_read_write(blobstore_two_docs):
    dd = blobstore_two_docs.as_dict()
    blobstore_two_docs.update([{"task_id": "mp-2", "data": dd}])
    res = blobstore_two_docs.query_one({"task_id": "mp-2"})
    assert res["data"]["@module"] == "maggma.stores.azure"


def test_remove(blobstore_two_docs):
    # At time of writing, Azurite does not support the delete_blobs operation
    # of the ContainerClient. See https://github.com/Azure/Azurite/issues/1809
    pytest.skip("Azurite currently does not support delete_blobs")

    def objects_in_bucket(key):
        objs = list(blobstore_two_docs.container.list_blobs())
        return key in [o["name"] for o in objs]

    blobstore_two_docs.update([{"task_id": "mp-2", "data": "asd"}])
    blobstore_two_docs.update([{"task_id": "mp-4", "data": "asd"}])
    blobstore_two_docs.update({"task_id": "mp-5", "data": "aaa"})
    assert blobstore_two_docs.query_one({"task_id": "mp-2"}) is not None
    assert blobstore_two_docs.query_one({"task_id": "mp-4"}) is not None
    assert objects_in_bucket("mp-2")
    assert objects_in_bucket("mp-4")

    blobstore_two_docs.remove_docs({"task_id": "mp-2"})
    blobstore_two_docs.remove_docs({"task_id": "mp-4"}, remove_blob_object=True)

    assert objects_in_bucket("mp-2")
    assert not objects_in_bucket("mp-4")

    assert blobstore_two_docs.query_one({"task_id": "mp-5"}) is not None


def test_close(blobstore_two_docs):
    list(blobstore_two_docs.query())
    blobstore_two_docs.close()
    with pytest.raises(RuntimeError):
        list(blobstore_two_docs.query())


def test_bad_import(mocker):
    mocker.patch("maggma.stores.azure.azure_blob", None)
    index = MemoryStore("index")
    with pytest.raises(RuntimeError):
        AzureBlobStore(index, "bucket1")


def test_eq(mongostore, blobstore_two_docs):
    assert blobstore_two_docs == blobstore_two_docs
    assert blobstore_two_docs != mongostore


def test_count_subdir(blobstore_w_subdir):
    blobstore_w_subdir.update([{"task_id": "mp-1", "data": "asd"}])
    blobstore_w_subdir.update([{"task_id": "mp-2", "data": "asd"}])

    assert blobstore_w_subdir.count() == 2
    assert blobstore_w_subdir.count({"task_id": "mp-2"}) == 1


def test_subdir_field(blobstore_w_subdir):
    blobstore_w_subdir.update([{"task_id": "mp-1", "data": "asd"}])
    blobstore_w_subdir.update([{"task_id": "mp-2", "data": "asd"}])

    for cc in blobstore_w_subdir.index.query():
        assert len(cc["sub_dir"]) > 0
        assert cc["sub_dir"] == blobstore_w_subdir.sub_dir


def test_remove_subdir(blobstore_w_subdir):
    blobstore_w_subdir.update([{"task_id": "mp-2", "data": "asd"}])
    blobstore_w_subdir.update([{"task_id": "mp-4", "data": "asd"}])

    assert blobstore_w_subdir.query_one({"task_id": "mp-2"}) is not None
    assert blobstore_w_subdir.query_one({"task_id": "mp-4"}) is not None

    blobstore_w_subdir.remove_docs({"task_id": "mp-2"})

    assert blobstore_w_subdir.query_one({"task_id": "mp-2"}) is None
    assert blobstore_w_subdir.query_one({"task_id": "mp-4"}) is not None


def test_searchable_fields(blobstore_two_docs):
    tic = datetime(2018, 4, 12, 16)

    data = [{"task_id": f"mp-{i}", "a": i, blobstore_two_docs.last_updated_field: tic} for i in range(4)]

    blobstore_two_docs.searchable_fields = ["task_id"]
    blobstore_two_docs.update(data, key="a")

    # This should only work if the searchable field was put into the index store
    assert set(blobstore_two_docs.distinct("task_id")) == {
        "mp-0",
        "mp-1",
        "mp-2",
        "mp-3",
    }


def test_newer_in(blobstore):
    tic = datetime(2018, 4, 12, 16)
    tic2 = datetime.utcnow()

    name_old = AZURITE_CONTAINER_NAME
    name_new = AZURITE_CONTAINER_NAME + "-2"
    with azurite_container(name_old), azurite_container(name_new):
        index_old = MemoryStore("index_old")
        old_store = AzureBlobStore(
            index_old,
            name_old,
            azure_client_info={"connection_string": AZURITE_CONNECTION_STRING},
        )
        old_store.connect()
        old_store.update([{"task_id": "mp-1", "last_updated": tic}])
        old_store.update([{"task_id": "mp-2", "last_updated": tic}])

        index_new = MemoryStore("index_new")
        new_store = AzureBlobStore(
            index_new,
            name_new,
            azure_client_info={"connection_string": AZURITE_CONNECTION_STRING},
        )
        new_store.connect()
        new_store.update([{"task_id": "mp-1", "last_updated": tic2}])
        new_store.update([{"task_id": "mp-2", "last_updated": tic2}])

    assert len(old_store.newer_in(new_store)) == 2
    assert len(new_store.newer_in(old_store)) == 0

    assert len(old_store.newer_in(new_store.index)) == 2
    assert len(new_store.newer_in(old_store.index)) == 0


def test_additional_metadata(blobstore_two_docs):
    tic = datetime(2018, 4, 12, 16)

    data = [{"task_id": f"mp-{i}", "a": i, blobstore_two_docs.last_updated_field: tic} for i in range(4)]

    blobstore_two_docs.update(data, key="a", additional_metadata="task_id")

    # This should only work if the searchable field was put into the index store
    assert set(blobstore_two_docs.distinct("task_id")) == {
        "mp-0",
        "mp-1",
        "mp-2",
        "mp-3",
    }


def test_no_container():
    with azurite_container(create_container=False):
        index = MemoryStore("index")
        store = AzureBlobStore(
            index,
            AZURITE_CONTAINER_NAME,
            azure_client_info={"connection_string": AZURITE_CONNECTION_STRING},
        )
        with pytest.raises(RuntimeError, match=r".*Container not present.*"):
            store.connect()

        # check that the store can create
        store = AzureBlobStore(
            index,
            AZURITE_CONTAINER_NAME,
            azure_client_info={"connection_string": AZURITE_CONNECTION_STRING},
            create_container=True,
        )
        store.connect()


def test_name(blobstore):
    assert blobstore.name == f"container://{AZURITE_CONTAINER_NAME}"


def test_ensure_index(blobstore_two_docs):
    assert blobstore_two_docs.ensure_index("task-id")


def test_no_login():
    with azurite_container():
        index = MemoryStore("index")
        store = AzureBlobStore(
            index,
            AZURITE_CONTAINER_NAME,
            azure_client_info={},
        )

    with pytest.raises(RuntimeError, match=r".*Could not instantiate BlobServiceClient.*"):
        store.connect()


def test_credential_type_valid():
    credential_type = "DefaultAzureCredential"
    index = MemoryStore("index")
    store = AzureBlobStore(
        index,
        AZURITE_CONTAINER_NAME,
        azure_client_info="client_url",
        credential_type=credential_type,
    )
    assert store.credential_type == credential_type

    # tricks the store into thinking you already
    # provided the blob service client so it skips
    # the connection checks. We are only testing that
    # the credential import works properly
    store.service = True
    store.connect()

    from azure.identity import DefaultAzureCredential

    credential_type = DefaultAzureCredential
    index = MemoryStore("index")
    store = AzureBlobStore(
        index,
        AZURITE_CONTAINER_NAME,
        azure_client_info="client_url",
        credential_type=credential_type,
    )
    assert not isinstance(store.credential_type, str)



from typing import Optional

from fastapi import Query
from fastapi.exceptions import HTTPException

from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS


class SortQuery(QueryOperator):
    """Method to generate the sorting portion of a query."""

    def __init__(self, fields: Optional[list[str]] = None, max_num: Optional[int] = None):
        """Sort query configuration.

        Args:
            fields (Optional[List[str]]): List of allowed fields to sort with
            max_num (Optional[int]): Max number of fields to simultaneously sort with

        """
        self.fields = fields or []
        self.max_num = max_num or 0

        if self.max_num < 0:
            raise ValueError("Max number of fields should be larger than 0")

    def query(
        self,
        _sort_fields: Optional[str] = Query(
            None,
            description="Comma delimited fields to sort with.\
 Prefixing '-' to a field will force a sort in descending order.",
        ),
    ) -> STORE_PARAMS:
        sort = {}

        if _sort_fields:
            field_list = _sort_fields.split(",")
            if self.max_num and len(field_list) > self.max_num:
                raise HTTPException(
                    status_code=400, detail=f"Please provide at most {self.max_num} field(s) to sort with"
                )

            for sort_field in field_list:
                query_entry = {sort_field: 1}

                if sort_field.startswith("-"):
                    query_entry = {sort_field[1:]: -1}
                    sort_field = sort_field[1:]

                if self.fields and sort_field not in self.fields:
                    continue

                sort.update(query_entry)

        return {"sort": sort}



import gzip
import pickle
from datetime import datetime, timedelta
from io import BytesIO, StringIO

import boto3
import jsonlines
import pandas as pd
import pytest
from bson import json_util
from moto import mock_aws

from maggma.stores.open_data import OpenDataStore, PandasMemoryStore, S3IndexStore, TasksOpenDataStore

pd.set_option("future.no_silent_downcasting", True)


# PandasMemoryStore tests
@pytest.fixture()
def memstore():
    store = PandasMemoryStore(key="task_id")
    store.update(
        pd.DataFrame(
            [
                {
                    store.key: "mp-1",
                    store.last_updated_field: datetime.utcnow(),
                    "data": "asd",
                    "int_val": 1,
                },
                {
                    store.key: "mp-3",
                    store.last_updated_field: datetime.utcnow(),
                    "data": "sdf",
                    "int_val": 3,
                },
            ]
        )
    )
    return store


def test_pdmems_pickle(memstore):
    sobj = pickle.dumps(memstore)
    dobj = pickle.loads(sobj)
    assert hash(dobj) == hash(memstore)
    assert dobj == memstore


def test_pdmems_query(memstore):
    # bad criteria
    with pytest.raises(AttributeError, match=r".*only support query or is_in"):
        memstore.query(criteria={"boo": "hoo"})
    with pytest.raises(AttributeError, match=r".*please just use one or the other"):
        memstore.query(
            criteria={
                "query": f"{memstore.key} == 'mp-1'",
                "is_in": ("data", ["sdf", "fdr"]),
                "boo": "hoo",
            }
        )
    # all
    pd = memstore.query()
    assert len(pd) == 2
    # query criteria
    pd = memstore.query(criteria={"query": f"{memstore.key} == 'mp-1'"})
    assert len(pd) == 1
    assert pd[memstore.key].iloc[0] == "mp-1"
    assert pd["data"].iloc[0] == "asd"
    # is_in criteria
    pd = memstore.query(criteria={"is_in": ("data", ["sdf", "fdr"]), "boo": "hoo"})
    assert len(pd) == 1
    assert pd[memstore.key].iloc[0] == "mp-3"
    assert pd["data"].iloc[0] == "sdf"
    # properties
    pd = memstore.query(properties=[memstore.key, memstore.last_updated_field])
    assert len(pd) == 2
    assert len(pd[memstore.key]) == 2
    with pytest.raises(KeyError):
        assert pd["data"]
    with pytest.raises(KeyError):
        memstore.query(properties=["fake"])
    # sort
    pd = memstore.query(sort={memstore.key: -1})
    assert len(pd) == 2
    assert pd[memstore.key].iloc[0] == "mp-3"
    # skip
    pd = memstore.query(sort={memstore.key: -1}, skip=1)
    assert len(pd) == 1
    assert pd[memstore.key].iloc[0] == "mp-1"
    # limit
    pd = memstore.query(sort={memstore.key: -1}, limit=1)
    assert len(pd) == 1
    assert pd[memstore.key].iloc[0] == "mp-3"
    # all
    pd = memstore.query(
        criteria={
            "is_in": ("data", ["sdf", "fdr", "asd"]),
            "boo": "hoo",
        },
        properties=[memstore.key],
        sort={"data": -1},
        skip=1,
        limit=1,
    )
    assert len(pd) == 1
    assert pd[memstore.key].iloc[0] == "mp-1"
    with pytest.raises(KeyError):
        assert pd["data"]


def test_pdmems_count(memstore):
    assert memstore.count() == 2
    assert memstore.count(criteria={"query": f"{memstore.key} == 'mp-1'"}) == 1
    assert memstore.count(criteria={"is_in": ("data", ["sdf", "fdr"]), "boo": "hoo"}) == 1
    assert PandasMemoryStore(key="task_id").count() == 0


@pytest.fixture()
def memstore2():
    store = PandasMemoryStore(key="task_id")
    store.update(
        pd.DataFrame(
            [
                {
                    store.key: "mp-1",
                    store.last_updated_field: datetime.utcnow() - timedelta(hours=1),
                    "data": "asd",
                },
                {
                    store.key: "mp-2",
                    store.last_updated_field: datetime.utcnow() + timedelta(hours=1),
                    "data": "asd",
                },
                {
                    store.key: "mp-3",
                    store.last_updated_field: datetime.utcnow() + timedelta(hours=1),
                    "data": "sdf",
                },
            ]
        )
    )
    return store


def test_pdmems_distinct(memstore2):
    assert len(memstore2.distinct(field=memstore2.key)) == len(memstore2._data)
    assert len(memstore2.distinct(field="data")) == 2
    assert len(memstore2.distinct(field=memstore2.key, criteria={"query": "data == 'asd'"})) == 2


def test_pdmems_last_updated(memstore2):
    assert memstore2._data[memstore2.last_updated_field].iloc[2] == memstore2.last_updated
    store = PandasMemoryStore(key="task_id")
    assert store.last_updated == datetime.min


def test_pdmems_newer_in(memstore, memstore2):
    s = memstore.newer_in(memstore2)
    assert len(s) == 2
    assert "mp-2" in s.unique()
    assert "mp-3" in s.unique()
    s = memstore.newer_in(target=memstore2, criteria={"query": "data == 'asd'"}, exhaustive=True)
    assert len(s) == 1
    assert "mp-2" in s.unique()
    with pytest.raises(AttributeError):
        memstore.newer_in(target=memstore2, criteria={"query": "data == 'asd'"}, exhaustive=False)


def test_pdmems_update(memstore):
    df = pd.DataFrame(
        [
            {
                memstore.key: "mp-1",
                memstore.last_updated_field: datetime.utcnow(),
                "data": "boo",
                "int_val": 1,
            }
        ]
    )
    df2 = memstore.update(df)
    assert len(memstore._data) == 2
    assert memstore.query(criteria={"query": f"{memstore.key} == 'mp-1'"})["data"].iloc[0] == "boo"
    assert memstore.query(criteria={"query": f"{memstore.key} == 'mp-1'"})["int_val"].iloc[0] == 1
    assert df2.equals(df)
    df = pd.DataFrame(
        [
            {
                memstore.key: "mp-2",
                memstore.last_updated_field: datetime.utcnow(),
                "data": "boo",
                "int_val": 2,
            }
        ]
    )
    df2 = memstore.update(df)
    assert len(memstore._data) == 3
    assert memstore.query(criteria={"query": f"{memstore.key} == 'mp-2'"})["data"].iloc[0] == "boo"
    assert memstore.query(criteria={"query": f"{memstore.key} == 'mp-2'"})["int_val"].iloc[0] == 2
    assert df2.equals(df)


@pytest.fixture()
def s3indexstore():
    data = [{"task_id": "mp-1", "last_updated": datetime.utcnow()}]
    with mock_aws():
        conn = boto3.resource("s3", region_name="us-east-1")
        conn.create_bucket(Bucket="bucket1")
        client = boto3.client("s3", region_name="us-east-1")
        string_io = StringIO()
        with jsonlines.Writer(string_io, dumps=json_util.dumps) as writer:
            for _, row in pd.DataFrame(data).iterrows():
                writer.write(row.to_dict())
        client.put_object(
            Bucket="bucket1",
            Body=BytesIO(string_io.getvalue().encode("utf-8")),
            Key="manifest.jsonl",
        )

        store = S3IndexStore(collection_name="index", bucket="bucket1", key="task_id")
        store.connect()

        yield store


def test_s3is_pickle(s3indexstore):
    sobj = pickle.dumps(s3indexstore)
    dobj = pickle.loads(sobj)
    assert hash(dobj) == hash(s3indexstore)
    assert dobj == s3indexstore


def test_s3is_connect_retrieve_manifest(s3indexstore):
    assert s3indexstore.retrieve_manifest().equals(s3indexstore._data)
    with mock_aws():
        with pytest.raises(s3indexstore.s3_client.exceptions.NoSuchBucket):
            S3IndexStore(collection_name="foo", bucket="bucket2", key="task_id").retrieve_manifest()
        with pytest.raises(RuntimeError):
            S3IndexStore(collection_name="foo", bucket="bucket2", key="task_id").connect()
        conn = boto3.resource("s3", region_name="us-east-1")
        conn.create_bucket(Bucket="bucket2")
        s3is = S3IndexStore(collection_name="foo", bucket="bucket2", key="task_id")
        s3is.connect()
        assert s3is._data is None
        assert s3is.retrieve_manifest() is None
        assert s3is.count() == 0


def test_s3is_store_manifest():
    with mock_aws():
        conn = boto3.resource("s3", region_name="us-east-1")
        conn.create_bucket(Bucket="bucket2")
        s3is = S3IndexStore(collection_name="foo", bucket="bucket2", key="task_id")
        s3is.connect()
        s3is.update(pd.DataFrame([{"task_id": "mp-2", "last_updated": "now"}]))
        s3is.store_manifest()
        df = s3is.retrieve_manifest()
        assert len(df) == 1
        assert df.equals(s3is._data)
        s3is.update(pd.DataFrame([{"task_id": "mp-3", "last_updated": "later"}]))
        df = s3is.retrieve_manifest()
        assert not df.equals(s3is._data)


def test_s3is_close(s3indexstore):
    s3indexstore.close()
    assert len(s3indexstore.query()) == 1  # actions auto-reconnect
    s3indexstore.update(pd.DataFrame([{"task_id": "mp-2", "last_updated": "now"}]))
    assert len(s3indexstore.query()) == 2
    s3indexstore.close()
    assert len(s3indexstore.query()) == 2  # actions auto-reconnect
    s3indexstore.close()
    s3indexstore.connect()
    assert len(s3indexstore.query()) == 1  # explicit connect reloads manifest


@pytest.fixture()
def s3store():
    with mock_aws():
        conn = boto3.resource("s3", region_name="us-east-1")
        conn.create_bucket(Bucket="bucket1")

        store = OpenDataStore(
            collection_name="index", bucket="bucket1", key="task_id", object_grouping=["group_level_two", "task_id"]
        )
        store.connect()

        store.update(
            pd.DataFrame(
                [
                    {
                        "task_id": "mp-1",
                        "data": "asd",
                        store.last_updated_field: datetime.utcnow(),
                        "group": {"level_two": 4},
                    },
                    {
                        "task_id": "mp-3",
                        "data": "sdf",
                        store.last_updated_field: datetime.utcnow(),
                        "group": {"level_two": 4},
                    },
                ]
            )
        )

        yield store


@pytest.fixture()
def s3store_w_subdir():
    with mock_aws():
        conn = boto3.resource("s3", region_name="us-east-1")
        conn.create_bucket(Bucket="bucket1")
        conn.create_bucket(Bucket="bucket2")

        index = S3IndexStore(collection_name="index", bucket="bucket1", key="task_id")
        store = OpenDataStore(
            index=index,
            collection_name=index.collection_name,
            bucket="bucket2",
            key=index.key,
            prefix="subdir1",
            object_grouping=["data_foo"],
        )
        store.connect()

        store.update(
            pd.DataFrame(
                [
                    {
                        "task_id": "mp-1",
                        "data": {"foo": "asd"},
                        store.last_updated_field: datetime.utcnow(),
                    },
                    {
                        "task_id": "mp-3",
                        "data": {"foo": "sdf"},
                        store.last_updated_field: datetime.utcnow(),
                    },
                ]
            )
        )

        yield store


def test_pickle(s3store):
    sobj = pickle.dumps(s3store)
    dobj = pickle.loads(sobj)
    assert hash(dobj) == hash(s3store)
    assert dobj == s3store
    assert len(dobj.query(criteria={"query": "task_id == 'mp-2'"})) == 0
    assert dobj.query(criteria={"query": "task_id == 'mp-1'"})["data"].iloc[0] == "asd"


def test_index_property(s3store, s3store_w_subdir):
    assert s3store.index != s3store
    assert s3store_w_subdir.index != s3store_w_subdir


def test_read_doc_from_s3():
    data = [
        {"task_id": "mp-1", "last_updated": datetime.utcnow(), "group": "foo", "also_group": "boo"},
        {"task_id": "mp-2", "last_updated": datetime.utcnow(), "group": "foo", "also_group": "boo"},
    ]
    with mock_aws():
        conn = boto3.resource("s3", region_name="us-east-1")
        conn.create_bucket(Bucket="bucket1")
        client = boto3.client("s3", region_name="us-east-1")
        string_io = StringIO()
        with jsonlines.Writer(string_io, dumps=json_util.dumps) as writer:
            for _, row in pd.DataFrame(data).iterrows():
                writer.write(row.to_dict())

        client.put_object(
            Bucket="bucket1",
            Body=BytesIO(gzip.compress(string_io.getvalue().encode("utf-8"))),
            Key="group=foo/also_group=boo.jsonl.gz",
        )

        store = OpenDataStore(
            collection_name="index", bucket="bucket1", key="task_id", object_grouping=["group", "also_group"]
        )
        store.connect()
        df = store._read_doc_from_s3(file_id="group=foo/also_group=boo.jsonl.gz")
        assert len(df) == 2
        assert store._get_full_key_path(index=df) == "group=foo/also_group=boo.jsonl.gz"
        assert (df["task_id"] == "mp-1").any()
        assert (df["task_id"] == "mp-2").any()


@pytest.mark.xfail(
    reason="Known issue, the store is in a deprecated state, and in particular may be incompatible with numpy 2.0+"
)
def test_update(s3store):
    assert len(s3store.index_data) == 2
    s3store.update(
        pd.DataFrame(
            [
                {
                    "task_id": "mp-199999",
                    "data": "asd",
                    "group": {"level_two": 4},
                    s3store.last_updated_field: datetime.utcnow(),
                }
            ]
        )
    )
    assert len(s3store.index_data) == 3
    with pytest.raises(KeyError):
        assert s3store.index_data.query("task_id == 'mp-199999'")["data"].iloc[0] == "asd"

    s3store.update(
        pd.DataFrame(
            [
                {
                    "task_id": "mp-199999",
                    "data": "foo",
                    "group": {"level_two": 4},
                    s3store.last_updated_field: datetime.utcnow(),
                }
            ]
        )
    )
    assert len(s3store.index_data) == 3
    assert len(s3store.index_data.query("task_id == 'mp-199999'")) == 1

    mp4 = [{"task_id": "mp-4", "data": "asd", "group": {"level_two": 4}, s3store.last_updated_field: datetime.utcnow()}]
    s3store.update(pd.DataFrame(mp4))
    assert len(s3store.index_data) == 4
    mp4_index = [{"task_id": "mp-4", "group_level_two": 4, s3store.last_updated_field: datetime.utcnow()}]
    assert s3store._get_full_key_path(pd.DataFrame(mp4_index)) == "group_level_two=4/task_id=mp-4.jsonl.gz"
    s3store.s3_client.head_object(Bucket=s3store.bucket, Key=s3store._get_full_key_path(pd.DataFrame(mp4_index)))


def test_query(s3store):
    assert len(s3store.query(criteria={"query": "task_id == 'mp-2'"})) == 0
    assert s3store.query(criteria={"query": "task_id == 'mp-1'"})["data"].iloc[0] == "asd"
    assert s3store.query(criteria={"query": "task_id == 'mp-3'"})["data"].iloc[0] == "sdf"
    assert s3store.query(criteria={"query": "task_id == 'mp-1'"}, properties=["task_id"])["task_id"].iloc[0] == "mp-1"
    assert (
        s3store.query(criteria={"query": "task_id == 'mp-1'"}, properties=["task_id", "data"])["data"].iloc[0] == "asd"
    )

    assert len(s3store.query()) == 2

    # will use optimized search
    df = s3store.query(criteria={"query": "task_id == 'mp-1'"}, properties=["task_id"], criteria_fields=["task_id"])
    assert len(df) == 1
    assert df["task_id"].iloc[0] == "mp-1"
    with pytest.raises(KeyError):
        assert df["data"].iloc[0] == "asd"

    # will use optimized search
    df = s3store.query(properties=["task_id"], criteria_fields=[])
    assert len(df) == 2

    # will not use optimized search even with hints since data is not in the searchable_fields
    df = s3store.query(criteria={"query": "data == 'asd'"}, properties=["task_id"], criteria_fields=["data"])
    assert len(df) == 1
    assert df["task_id"].iloc[0] == "mp-1"
    with pytest.raises(KeyError):
        assert df["data"].iloc[0] == "asd"


def test_rebuild_index_from_s3_data(s3store):
    data = [
        {"task_id": "mp-2", "data": "asd", s3store.last_updated_field: datetime.utcnow(), "group": {"level_two": 4}}
    ]
    client = boto3.client("s3", region_name="us-east-1")
    string_io = StringIO()
    with jsonlines.Writer(string_io, dumps=json_util.dumps) as writer:
        for _, row in pd.DataFrame(data).iterrows():
            writer.write(row.to_dict())

    data = [
        {"task_id": "mp-99", "data": "asd", s3store.last_updated_field: datetime.utcnow(), "group": {"level_two": 4}}
    ]
    string_io2 = StringIO()
    with jsonlines.Writer(string_io2, dumps=json_util.dumps) as writer:
        for _, row in pd.DataFrame(data).iterrows():
            writer.write(row.to_dict())

    client.put_object(
        Bucket="bucket1",
        Body=BytesIO(gzip.compress(string_io.getvalue().encode("utf-8"))),
        Key="group_level_two=4/task_id=mp-2.jsonl.gz",
    )

    # creating file that should not be indexed to test that it gets skipped
    client.put_object(
        Bucket="bucket1",
        Body=BytesIO(gzip.compress(string_io2.getvalue().encode("utf-8"))),
        Key="task_id=mp-99.gz",
    )

    assert len(s3store.index.index_data) == 2
    index_docs = s3store.rebuild_index_from_s3_data()
    assert len(index_docs) == 3
    assert len(s3store.index.index_data) == 3
    for key in index_docs.columns:
        assert key == "task_id" or key == "last_updated" or key == "group_level_two"


def test_rebuild_index_from_data(s3store):
    data = [
        {"task_id": "mp-2", "data": "asd", s3store.last_updated_field: datetime.utcnow(), "group": {"level_two": 4}}
    ]
    index_docs = s3store.rebuild_index_from_data(pd.DataFrame(data))
    assert len(index_docs) == 1
    assert len(s3store.index.index_data) == 1
    for key in index_docs.columns:
        assert key == "task_id" or key == "last_updated" or key == "group_level_two"


def test_count_subdir(s3store_w_subdir):
    s3store_w_subdir.update(
        pd.DataFrame(
            [{"task_id": "mp-1", "data": {"foo": "asd"}, s3store_w_subdir.last_updated_field: datetime.utcnow()}]
        )
    )
    s3store_w_subdir.update(
        pd.DataFrame(
            [{"task_id": "mp-2", "data": {"foo": "asd"}, s3store_w_subdir.last_updated_field: datetime.utcnow()}]
        )
    )

    assert len(s3store_w_subdir.query()) == 3
    assert len(s3store_w_subdir.query(criteria=None)) == 3
    assert s3store_w_subdir.count() == 3
    assert s3store_w_subdir.count({"query": "task_id == 'mp-2'"}) == 1


def test_subdir_storage(s3store_w_subdir):
    def objects_in_bucket(key):
        objs = s3store_w_subdir.s3_client.list_objects_v2(Bucket=s3store_w_subdir.bucket, Prefix=key)
        return key in [o["Key"] for o in objs["Contents"]]

    s3store_w_subdir.update(
        pd.DataFrame(
            [{"task_id": "mp-1", "data": {"foo": "asd"}, s3store_w_subdir.last_updated_field: datetime.utcnow()}]
        )
    )
    s3store_w_subdir.update(
        pd.DataFrame(
            [{"task_id": "mp-2", "data": {"foo": "asd"}, s3store_w_subdir.last_updated_field: datetime.utcnow()}]
        )
    )

    assert objects_in_bucket("subdir1/data_foo=asd.jsonl.gz")
    assert objects_in_bucket("subdir1/data_foo=sdf.jsonl.gz")


def test_additional_metadata(s3store):
    tic = datetime(2018, 4, 12, 16)

    data = [{"task_id": f"mp-{i}", "a": i, s3store.last_updated_field: tic} for i in range(4)]

    with pytest.raises(TypeError):
        s3store.update(data, key="a", additional_metadata="task_id")


def test_rebuild_index_from_s3_for_tasks_store():
    data = [{"task_id": "mp-2", "data": "asd", "last_updated": datetime.utcnow(), "group": {"level_two": 4}}]
    with mock_aws():
        conn = boto3.resource("s3", region_name="us-east-1")
        conn.create_bucket(Bucket="bucket1")

        string_io = StringIO()
        with jsonlines.Writer(string_io, dumps=json_util.dumps) as writer:
            for _, row in pd.DataFrame(data).iterrows():
                writer.write(row.to_dict())

        client = boto3.client("s3", region_name="us-east-1")
        client.put_object(
            Bucket="bucket1",
            Body=BytesIO(gzip.compress(string_io.getvalue().encode("utf-8"))),
            Key="group_level_two=4/dt=some_random_data.jsonl.gz",
        )

        store = TasksOpenDataStore(
            collection_name="index", bucket="bucket1", key="task_id", object_grouping=["group_level_two", "dt"]
        )
        store.connect()

        index_docs = store.rebuild_index_from_s3_data()
        assert len(index_docs) == 1
        assert len(store.index.index_data) == 1
        for key in index_docs.columns:
            assert key == "task_id" or key == "last_updated" or key == "group_level_two" or key == "dt"
        assert index_docs["dt"].iloc[0] == "some_random_data"


def test_no_update_for_tasks_store():
    with mock_aws():
        conn = boto3.resource("s3", region_name="us-east-1")
        conn.create_bucket(Bucket="bucket1")
        store = TasksOpenDataStore(
            collection_name="index", bucket="bucket1", key="task_id", object_grouping=["group_level_two", "dt"]
        )
        store.connect()

        with pytest.raises(NotImplementedError):
            store.update(
                pd.DataFrame(
                    [
                        {
                            "task_id": "mp-199999",
                            "data": "foo",
                            "group": {"level_two": 4},
                            store.last_updated_field: datetime.utcnow(),
                        }
                    ]
                )
            )
